from autolens.plotters import array_plotters
from autolens.plotters import plotter_util
from autolens.plotters import quantity_radii_plotters


def plot_image(
    light_profile,
    grid,
    mask=None,
    extract_array_from_mask=False,
    zoom_around_mask=False,
    positions=None,
    as_subplot=False,
    units="arcsec",
    kpc_per_arcsec=None,
    figsize=(7, 7),
    aspect="square",
    cmap="jet",
    norm="linear",
    norm_min=None,
    norm_max=None,
    linthresh=0.05,
    linscale=0.01,
    cb_ticksize=10,
    cb_fraction=0.047,
    cb_pad=0.01,
    cb_tick_values=None,
    cb_tick_labels=None,
    title="Image",
    titlesize=16,
    xlabelsize=16,
    ylabelsize=16,
    xyticksize=16,
    mask_pointsize=10,
    position_pointsize=10.0,
    grid_pointsize=1,
    output_path=None,
    output_format="show",
    output_filename="image",
):
    """Plot the image of a light profile, on a grid of (y,x) coordinates.

    Set *autolens.hyper_galaxies.array.plotters.array_plotters* for a description of all innput parameters not described below.

    Parameters
    -----------
    light_profile : model.profiles.light_profiles.LightProfile
        The light profile whose image are plotted.
    grid : ndarray or hyper_galaxies.array.grid_stacks.Grid
        The (y,x) coordinates of the grid, in an array of shape (total_coordinates, 2)
    """
    image = light_profile.profile_image_from_grid(
        grid=grid, return_in_2d=True, return_binned=True
    )

    array_plotters.plot_array(
        array=image,
        mask=mask,
        extract_array_from_mask=extract_array_from_mask,
        zoom_around_mask=zoom_around_mask,
        positions=positions,
        as_subplot=as_subplot,
        units=units,
        kpc_per_arcsec=kpc_per_arcsec,
        figsize=figsize,
        aspect=aspect,
        cmap=cmap,
        norm=norm,
        norm_min=norm_min,
        norm_max=norm_max,
        linthresh=linthresh,
        linscale=linscale,
        cb_ticksize=cb_ticksize,
        cb_fraction=cb_fraction,
        cb_pad=cb_pad,
        cb_tick_values=cb_tick_values,
        cb_tick_labels=cb_tick_labels,
        title=title,
        titlesize=titlesize,
        xlabelsize=xlabelsize,
        ylabelsize=ylabelsize,
        xyticksize=xyticksize,
        mask_pointsize=mask_pointsize,
        position_pointsize=position_pointsize,
        grid_pointsize=grid_pointsize,
        output_path=output_path,
        output_format=output_format,
        output_filename=output_filename,
    )


def plot_luminosity_within_circle_in_electrons_per_second_as_function_of_radius(
    light_profile,
    minimum_radius=1.0e-4,
    maximum_radius=10.0,
    radii_bins=10,
    as_subplot=False,
    label="Light Profile",
    plot_axis_type="semilogy",
    effective_radius_line=None,
    einstein_radius_line=None,
    units="arcsec",
    kpc_per_arcsec=None,
    figsize=(7, 7),
    plot_legend=True,
    title="Luminosity (Electrons Per Second) vs Radius",
    ylabel="Luminosity (Electrons Per Second)",
    titlesize=16,
    xlabelsize=16,
    ylabelsize=16,
    xyticksize=16,
    legend_fontsize=12,
    output_path=None,
    output_format="show",
    output_filename="luminosity_vs_radius",
):

    radii = plotter_util.quantity_radii_from_minimum_and_maximum_radii_and_radii_points(
        minimum_radius=minimum_radius,
        maximum_radius=maximum_radius,
        radii_points=radii_bins,
    )

    luminosities = list(
        map(
            lambda radius: light_profile.luminosity_within_circle_in_units(
                radius=radius
            ),
            radii,
        )
    )

    quantity_radii_plotters.plot_quantity_as_function_of_radius(
        quantity=luminosities,
        radii=radii,
        as_subplot=as_subplot,
        label=label,
        plot_axis_type=plot_axis_type,
        effective_radius_line=effective_radius_line,
        einstein_radius_line=einstein_radius_line,
        units=units,
        kpc_per_arcsec=kpc_per_arcsec,
        figsize=figsize,
        plot_legend=plot_legend,
        title=title,
        ylabel=ylabel,
        titlesize=titlesize,
        xlabelsize=xlabelsize,
        ylabelsize=ylabelsize,
        xyticksize=xyticksize,
        legend_fontsize=legend_fontsize,
        output_path=output_path,
        output_format=output_format,
        output_filename=output_filename,
    )


def plot_convergence(
    mass_profile,
    grid,
    mask=None,
    extract_array_from_mask=False,
    zoom_around_mask=False,
    positions=None,
    plot_critical_curves=False,
    plot_caustics=False,
    as_subplot=False,
    units="arcsec",
    kpc_per_arcsec=None,
    figsize=(7, 7),
    aspect="square",
    cmap="jet",
    norm="linear",
    norm_min=None,
    norm_max=None,
    linthresh=0.05,
    linscale=0.01,
    cb_ticksize=10,
    cb_fraction=0.047,
    cb_pad=0.01,
    cb_tick_values=None,
    cb_tick_labels=None,
    title="Convergence",
    titlesize=16,
    xlabelsize=16,
    ylabelsize=16,
    xyticksize=16,
    mask_pointsize=10,
    position_pointsize=10.0,
    grid_pointsize=1,
    output_path=None,
    output_format="show",
    output_filename="convergence",
):
    """Plot the convergence of a mass profile, on a grid of (y,x) coordinates.

    Set *autolens.hyper_galaxies.array.plotters.array_plotters* for a description of all innput parameters not described below.

    Parameters
    -----------
    mass_profile : model.profiles.mass_profiles.MassProfile
        The mass profile whose convergence is plotted.
    grid : ndarray or hyper_galaxies.array.grid_stacks.Grid
        The (y,x) coordinates of the grid, in an array of shape (total_coordinates, 2)
    """

    convergence = mass_profile.convergence_from_grid(
        grid=grid, return_in_2d=True, return_binned=True
    )

    lines = plotter_util.get_critical_curve_and_caustic(
        obj=mass_profile,
        grid=grid,
        plot_critical_curve=plot_critical_curves,
        plot_caustics=plot_caustics,
    )

    array_plotters.plot_array(
        array=convergence,
        mask=mask,
        extract_array_from_mask=extract_array_from_mask,
        zoom_around_mask=zoom_around_mask,
        positions=positions,
        lines=lines,
        as_subplot=as_subplot,
        units=units,
        kpc_per_arcsec=kpc_per_arcsec,
        figsize=figsize,
        aspect=aspect,
        cmap=cmap,
        norm=norm,
        norm_min=norm_min,
        norm_max=norm_max,
        linthresh=linthresh,
        linscale=linscale,
        cb_ticksize=cb_ticksize,
        cb_fraction=cb_fraction,
        cb_pad=cb_pad,
        cb_tick_values=cb_tick_values,
        cb_tick_labels=cb_tick_labels,
        title=title,
        titlesize=titlesize,
        xlabelsize=xlabelsize,
        ylabelsize=ylabelsize,
        xyticksize=xyticksize,
        mask_pointsize=mask_pointsize,
        position_pointsize=position_pointsize,
        grid_pointsize=grid_pointsize,
        output_path=output_path,
        output_format=output_format,
        output_filename=output_filename,
    )


def plot_potential(
    mass_profile,
    grid,
    mask=None,
    extract_array_from_mask=False,
    zoom_around_mask=False,
    positions=None,
    as_subplot=False,
    plot_critical_curves=False,
    plot_caustics=False,
    units="arcsec",
    kpc_per_arcsec=None,
    figsize=(7, 7),
    aspect="square",
    cmap="jet",
    norm="linear",
    norm_min=None,
    norm_max=None,
    linthresh=0.05,
    linscale=0.01,
    cb_ticksize=10,
    cb_fraction=0.047,
    cb_pad=0.01,
    cb_tick_values=None,
    cb_tick_labels=None,
    title="Potential",
    titlesize=16,
    xlabelsize=16,
    ylabelsize=16,
    xyticksize=16,
    mask_pointsize=10,
    position_pointsize=10.0,
    grid_pointsize=1,
    output_path=None,
    output_format="show",
    output_filename="potential",
):
    """Plot the potential of a mass profile, on a grid of (y,x) coordinates.

    Set *autolens.hyper_galaxies.array.plotters.array_plotters* for a description of all innput parameters not described below.

    Parameters
    -----------
    mass_profile : model.profiles.mass_profiles.MassProfile
        The mass profile whose potential is plotted.
    grid : ndarray or hyper_galaxies.array.grid_stacks.Grid
        The (y,x) coordinates of the grid, in an array of shape (total_coordinates, 2)
    """
    potential = mass_profile.potential_from_grid(
        grid=grid, return_in_2d=True, return_binned=True
    )

    lines = plotter_util.get_critical_curve_and_caustic(
        obj=mass_profile,
        grid=grid,
        plot_critical_curve=plot_critical_curves,
        plot_caustics=plot_caustics,
    )

    array_plotters.plot_array(
        array=potential,
        mask=mask,
        extract_array_from_mask=extract_array_from_mask,
        zoom_around_mask=zoom_around_mask,
        positions=positions,
        lines=lines,
        as_subplot=as_subplot,
        units=units,
        kpc_per_arcsec=kpc_per_arcsec,
        figsize=figsize,
        aspect=aspect,
        cmap=cmap,
        norm=norm,
        norm_min=norm_min,
        norm_max=norm_max,
        linthresh=linthresh,
        linscale=linscale,
        cb_ticksize=cb_ticksize,
        cb_fraction=cb_fraction,
        cb_pad=cb_pad,
        cb_tick_values=cb_tick_values,
        cb_tick_labels=cb_tick_labels,
        title=title,
        titlesize=titlesize,
        xlabelsize=xlabelsize,
        ylabelsize=ylabelsize,
        xyticksize=xyticksize,
        mask_pointsize=mask_pointsize,
        position_pointsize=position_pointsize,
        grid_pointsize=grid_pointsize,
        output_path=output_path,
        output_format=output_format,
        output_filename=output_filename,
    )


def plot_deflections_y(
    mass_profile,
    grid,
    mask=None,
    extract_array_from_mask=False,
    zoom_around_mask=False,
    positions=None,
    plot_critical_curves=False,
    plot_caustics=False,
    as_subplot=False,
    units="arcsec",
    kpc_per_arcsec=None,
    figsize=(7, 7),
    aspect="square",
    cmap="jet",
    norm="linear",
    norm_min=None,
    norm_max=None,
    linthresh=0.05,
    linscale=0.01,
    cb_ticksize=10,
    cb_fraction=0.047,
    cb_pad=0.01,
    cb_tick_values=None,
    cb_tick_labels=None,
    title="Deflections (y)",
    titlesize=16,
    xlabelsize=16,
    ylabelsize=16,
    xyticksize=16,
    mask_pointsize=10,
    position_pointsize=10.0,
    grid_pointsize=1,
    output_path=None,
    output_format="show",
    output_filename="deflections_y",
):
    """Plot the y component of the deflection angles of a mass profile, on a grid of (y,x) coordinates.

    Set *autolens.hyper_galaxies.array.plotters.array_plotters* for a description of all innput parameters not described below.

    Parameters
    -----------
    mass_profile : model.profiles.mass_profiles.MassProfile
        The mass profile whose y deflecton angles are plotted.
    grid : ndarray or hyper_galaxies.array.grid_stacks.Grid
        The (y,x) coordinates of the grid, in an array of shape (total_coordinates, 2)
    """

    deflections = mass_profile.deflections_from_grid(
        grid, return_in_2d=False, return_binned=True
    )
    deflections_y = grid.scaled_array_2d_from_array_1d(deflections[:, 0])

    lines = plotter_util.get_critical_curve_and_caustic(
        obj=mass_profile,
        grid=grid,
        plot_critical_curve=plot_critical_curves,
        plot_caustics=plot_caustics,
    )

    array_plotters.plot_array(
        array=deflections_y,
        mask=mask,
        extract_array_from_mask=extract_array_from_mask,
        zoom_around_mask=zoom_around_mask,
        positions=positions,
        lines=lines,
        as_subplot=as_subplot,
        units=units,
        kpc_per_arcsec=kpc_per_arcsec,
        figsize=figsize,
        aspect=aspect,
        cmap=cmap,
        norm=norm,
        norm_min=norm_min,
        norm_max=norm_max,
        linthresh=linthresh,
        linscale=linscale,
        cb_ticksize=cb_ticksize,
        cb_fraction=cb_fraction,
        cb_pad=cb_pad,
        cb_tick_values=cb_tick_values,
        cb_tick_labels=cb_tick_labels,
        title=title,
        titlesize=titlesize,
        xlabelsize=xlabelsize,
        ylabelsize=ylabelsize,
        xyticksize=xyticksize,
        mask_pointsize=mask_pointsize,
        position_pointsize=position_pointsize,
        grid_pointsize=grid_pointsize,
        output_path=output_path,
        output_format=output_format,
        output_filename=output_filename,
    )


def plot_deflections_x(
    mass_profile,
    grid,
    mask=None,
    extract_array_from_mask=False,
    zoom_around_mask=False,
    positions=None,
    plot_critical_curves=False,
    plot_caustics=False,
    as_subplot=False,
    units="arcsec",
    kpc_per_arcsec=None,
    figsize=(7, 7),
    aspect="square",
    cmap="jet",
    norm="linear",
    norm_min=None,
    norm_max=None,
    linthresh=0.05,
    linscale=0.01,
    cb_ticksize=10,
    cb_fraction=0.047,
    cb_pad=0.01,
    cb_tick_values=None,
    cb_tick_labels=None,
    title="Deflections (x)",
    titlesize=16,
    xlabelsize=16,
    ylabelsize=16,
    xyticksize=16,
    mask_pointsize=10,
    position_pointsize=10.0,
    grid_pointsize=1,
    output_path=None,
    output_format="show",
    output_filename="deflections_x",
):
    """Plot the x component of the deflection angles of a mass profile, on a grid of (y,x) coordinates.

     Set *autolens.hyper_galaxies.array.plotters.array_plotters* for a description of all innput parameters not described below.

     Parameters
     -----------
     mass_profile : model.profiles.mass_profiles.MassProfile
         The mass profile whose x deflecton angles are plotted.
     grid : ndarray or hyper_galaxies.array.grid_stacks.Grid
         The (y,x) coordinates of the grid, in an array of shape (total_coordinates, 2)
     """
    deflections = mass_profile.deflections_from_grid(
        grid, return_in_2d=False, return_binned=True
    )
    deflections_x = grid.scaled_array_2d_from_array_1d(deflections[:, 1])

    lines = plotter_util.get_critical_curve_and_caustic(
        obj=mass_profile,
        grid=grid,
        plot_critical_curve=plot_critical_curves,
        plot_caustics=plot_caustics,
    )

    array_plotters.plot_array(
        array=deflections_x,
        mask=mask,
        extract_array_from_mask=extract_array_from_mask,
        zoom_around_mask=zoom_around_mask,
        positions=positions,
        lines=lines,
        as_subplot=as_subplot,
        units=units,
        kpc_per_arcsec=kpc_per_arcsec,
        figsize=figsize,
        aspect=aspect,
        cmap=cmap,
        norm=norm,
        norm_min=norm_min,
        norm_max=norm_max,
        linthresh=linthresh,
        linscale=linscale,
        cb_ticksize=cb_ticksize,
        cb_fraction=cb_fraction,
        cb_pad=cb_pad,
        cb_tick_values=cb_tick_values,
        cb_tick_labels=cb_tick_labels,
        title=title,
        titlesize=titlesize,
        xlabelsize=xlabelsize,
        ylabelsize=ylabelsize,
        xyticksize=xyticksize,
        mask_pointsize=mask_pointsize,
        position_pointsize=position_pointsize,
        grid_pointsize=grid_pointsize,
        output_path=output_path,
        output_format=output_format,
        output_filename=output_filename,
    )


def plot_magnification(
    mass_profile,
    grid,
    mask=None,
    extract_array_from_mask=False,
    zoom_around_mask=False,
    positions=None,
    as_subplot=False,
    plot_critical_curves=False,
    plot_caustics=False,
    units="arcsec",
    kpc_per_arcsec=None,
    figsize=(7, 7),
    aspect="square",
    cmap="jet",
    norm="linear",
    norm_min=None,
    norm_max=None,
    linthresh=0.05,
    linscale=0.01,
    cb_ticksize=10,
    cb_fraction=0.047,
    cb_pad=0.01,
    cb_tick_values=None,
    cb_tick_labels=None,
    title="Magnification",
    titlesize=16,
    xlabelsize=16,
    ylabelsize=16,
    xyticksize=16,
    mask_pointsize=10,
    position_pointsize=10.0,
    grid_pointsize=1,
    output_path=None,
    output_format="show",
    output_filename="magnification",
):
    """Plot the magnification of a mass profile, on a grid of (y,x) coordinates.

    Set *autolens.hyper_galaxies.array.plotters.array_plotters* for a description of all innput parameters not described below.

    Parameters
    -----------
    mass_profile : model.profiles.mass_profiles.MassProfile
        The mass profile whose magnification is plotted.
    grid : ndarray or hyper_galaxies.array.grid_stacks.Grid
        The (y,x) coordinates of the grid, in an array of shape (total_coordinates, 2)
    """
    magnification = mass_profile.magnification_from_grid(
        grid=grid, return_in_2d=True, return_binned=True
    )

    lines = plotter_util.get_critical_curve_and_caustic(
        obj=mass_profile,
        grid=grid,
        plot_critical_curve=plot_critical_curves,
        plot_caustics=plot_caustics,
    )

    array_plotters.plot_array(
        array=magnification,
        mask=mask,
        extract_array_from_mask=extract_array_from_mask,
        zoom_around_mask=zoom_around_mask,
        positions=positions,
        lines=lines,
        as_subplot=as_subplot,
        units=units,
        kpc_per_arcsec=kpc_per_arcsec,
        figsize=figsize,
        aspect=aspect,
        cmap=cmap,
        norm=norm,
        norm_min=norm_min,
        norm_max=norm_max,
        linthresh=linthresh,
        linscale=linscale,
        cb_ticksize=cb_ticksize,
        cb_fraction=cb_fraction,
        cb_pad=cb_pad,
        cb_tick_values=cb_tick_values,
        cb_tick_labels=cb_tick_labels,
        title=title,
        titlesize=titlesize,
        xlabelsize=xlabelsize,
        ylabelsize=ylabelsize,
        xyticksize=xyticksize,
        mask_pointsize=mask_pointsize,
        position_pointsize=position_pointsize,
        grid_pointsize=grid_pointsize,
        output_path=output_path,
        output_format=output_format,
        output_filename=output_filename,
    )



import autolens as al
from autolens.data import observation as obs

import numpy as np
import os
import shutil

test_data_dir = "{}/../test_files/array/".format(
    os.path.dirname(os.path.realpath(__file__))
)


class TestObservation:
    def test__constructor_and_specific_instrument_class_methods(self):

        psf = al.PSF.from_gaussian(shape=(11, 11), sigma=0.1, pixel_scale=0.1)

        observation = obs.CCDObservation(
            shape=(51, 51),
            pixel_scale=0.1,
            psf=psf,
            exposure_time=20.0,
            background_sky_level=10.0,
        )

        assert observation.shape == (51, 51)
        assert observation.pixel_scale == 0.1
        assert observation.psf == psf
        assert observation.exposure_time == 20.0
        assert observation.background_sky_level == 10.0

        lsst = obs.CCDObservation.lsst()

        lsst_psf = al.PSF.from_gaussian(shape=(31, 31), sigma=0.5, pixel_scale=0.2)

        assert lsst.shape == (101, 101)
        assert lsst.pixel_scale == 0.2
        assert lsst.psf == lsst_psf
        assert lsst.exposure_time == 100.0
        assert lsst.background_sky_level == 1.0

        euclid = obs.CCDObservation.euclid()

        euclid_psf = al.PSF.from_gaussian(shape=(31, 31), sigma=0.1, pixel_scale=0.1)

        assert euclid.shape == (151, 151)
        assert euclid.pixel_scale == 0.1
        assert euclid.psf == euclid_psf
        assert euclid.exposure_time == 565.0
        assert euclid.background_sky_level == 1.0

        hst = obs.CCDObservation.hst()

        hst_psf = al.PSF.from_gaussian(shape=(31, 31), sigma=0.05, pixel_scale=0.05)

        assert hst.shape == (251, 251)
        assert hst.pixel_scale == 0.05
        assert hst.psf == hst_psf
        assert hst.exposure_time == 2000.0
        assert hst.background_sky_level == 1.0

        hst_up_sampled = obs.CCDObservation.hst_up_sampled()

        hst_up_sampled_psf = al.PSF.from_gaussian(
            shape=(31, 31), sigma=0.05, pixel_scale=0.03
        )

        assert hst_up_sampled.shape == (401, 401)
        assert hst_up_sampled.pixel_scale == 0.03
        assert hst_up_sampled.psf == hst_up_sampled_psf
        assert hst_up_sampled.exposure_time == 2000.0
        assert hst_up_sampled.background_sky_level == 1.0

        adaptive_optics = obs.CCDObservation.keck_adaptive_optics()

        adaptive_optics_psf = al.PSF.from_gaussian(
            shape=(31, 31), sigma=0.025, pixel_scale=0.01
        )

        assert adaptive_optics.shape == (751, 751)
        assert adaptive_optics.pixel_scale == 0.01
        assert adaptive_optics.psf == adaptive_optics_psf
        assert adaptive_optics.exposure_time == 1000.0
        assert adaptive_optics.background_sky_level == 1.0

    def test__simulate_ccd_data_from_lens__source_galaxy__compare_to_manual_ccd_data(
        self
    ):

        lens_galaxy = al.Galaxy(
            redshift=0.5,
            mass=al.mass_profiles.EllipticalIsothermal(
                centre=(0.0, 0.0), einstein_radius=1.6, axis_ratio=0.7, phi=45.0
            ),
        )

        source_galaxy = al.Galaxy(
            redshift=0.5,
            light=al.light_profiles.EllipticalSersic(
                centre=(0.1, 0.1),
                axis_ratio=0.8,
                phi=60.0,
                intensity=0.3,
                effective_radius=1.0,
                sersic_index=2.5,
            ),
        )

        grid = al.Grid.from_shape_pixel_scale_and_sub_grid_size(
            shape=(11, 11), pixel_scale=0.2, sub_grid_size=1
        )

        tracer = al.Tracer.from_galaxies(galaxies=[lens_galaxy, source_galaxy])

        shape = (11, 11)
        pixel_scale = 0.2
        psf = al.PSF.from_gaussian(shape=(7, 7), sigma=0.1, pixel_scale=0.2)
        exposure_time = 100.0
        background_sky_level = 1.0

        ccd_data = al.SimulatedCCDData.from_tracer_grid_and_exposure_arrays(
            tracer=tracer,
            grid=grid,
            pixel_scale=pixel_scale,
            exposure_time=exposure_time,
            psf=psf,
            background_sky_level=background_sky_level,
            add_noise=False,
            noise_if_add_noise_false=0.2,
        )

        observation = obs.CCDObservation(
            shape=shape,
            pixel_scale=pixel_scale,
            psf=psf,
            exposure_time=exposure_time,
            background_sky_level=background_sky_level,
        )

        observation_ccd_data = observation.simulate_ccd_data_from_galaxies(
            galaxies=[lens_galaxy, source_galaxy],
            sub_grid_size=1,
            add_noise=False,
            noise_if_add_noise_false=0.2,
        )

        assert (ccd_data.image == observation_ccd_data.image).all()
        assert (ccd_data.psf == observation_ccd_data.psf).all()
        assert observation_ccd_data.noise_map == 0.2 * np.ones((11, 11))
        assert ccd_data.noise_map == observation_ccd_data.noise_map
        assert (
            ccd_data.background_noise_map == observation_ccd_data.background_noise_map
        )
        assert ccd_data.poisson_noise_map == observation_ccd_data.poisson_noise_map
        assert (
            ccd_data.exposure_time_map == observation_ccd_data.exposure_time_map
        ).all()
        assert (
            ccd_data.background_sky_map == observation_ccd_data.background_sky_map
        ).all()

        ccd_data = al.SimulatedCCDData.from_tracer_grid_and_exposure_arrays(
            tracer=tracer,
            grid=grid,
            pixel_scale=pixel_scale,
            exposure_time=exposure_time,
            psf=psf,
            background_sky_level=background_sky_level,
            add_noise=True,
            noise_seed=1,
        )

        observation_ccd_data = observation.simulate_ccd_data_from_galaxies(
            galaxies=[lens_galaxy, source_galaxy],
            sub_grid_size=1,
            add_noise=True,
            noise_seed=1,
        )

        assert (ccd_data.image == observation_ccd_data.image).all()
        assert (ccd_data.psf == observation_ccd_data.psf).all()
        assert (ccd_data.noise_map == observation_ccd_data.noise_map).all()
        assert (
            ccd_data.background_noise_map == observation_ccd_data.background_noise_map
        ).all()
        assert (
            ccd_data.poisson_noise_map == observation_ccd_data.poisson_noise_map
        ).all()
        assert (
            ccd_data.exposure_time_map == observation_ccd_data.exposure_time_map
        ).all()
        assert (
            ccd_data.background_sky_map == observation_ccd_data.background_sky_map
        ).all()

    def test__simulate_ccd_data_from_lens__source_galaxy__and_write_to_fits(self):

        lens_galaxy = al.Galaxy(
            redshift=0.5,
            mass=al.mass_profiles.EllipticalIsothermal(
                centre=(0.0, 0.0), einstein_radius=1.6, axis_ratio=0.7, phi=45.0
            ),
        )

        source_galaxy = al.Galaxy(
            redshift=0.5,
            light=al.light_profiles.EllipticalSersic(
                centre=(0.1, 0.1),
                axis_ratio=0.8,
                phi=60.0,
                intensity=0.3,
                effective_radius=1.0,
                sersic_index=2.5,
            ),
        )

        grid = al.Grid.from_shape_pixel_scale_and_sub_grid_size(
            shape=(11, 11), pixel_scale=0.2, sub_grid_size=1
        )

        tracer = al.Tracer.from_galaxies(galaxies=[lens_galaxy, source_galaxy])

        shape = (11, 11)
        pixel_scale = 0.2
        psf = al.PSF.from_gaussian(shape=(7, 7), sigma=0.1, pixel_scale=0.2)
        exposure_time = 100.0
        background_sky_level = 1.0

        ccd_data = al.SimulatedCCDData.from_tracer_grid_and_exposure_arrays(
            tracer=tracer,
            grid=grid,
            pixel_scale=pixel_scale,
            exposure_time=exposure_time,
            psf=psf,
            background_sky_level=background_sky_level,
            add_noise=False,
            noise_if_add_noise_false=0.2,
        )

        observation = obs.CCDObservation(
            shape=shape,
            pixel_scale=pixel_scale,
            psf=psf,
            exposure_time=exposure_time,
            background_sky_level=background_sky_level,
        )

        output_data_dir = "{}/../test_files/array/output_test/".format(
            os.path.dirname(os.path.realpath(__file__))
        )
        if os.path.exists(output_data_dir):
            shutil.rmtree(output_data_dir)

        os.makedirs(output_data_dir)

        observation.simulate_ccd_data_from_galaxies_and_write_to_fits(
            galaxies=[lens_galaxy, source_galaxy],
            data_path=output_data_dir,
            data_name="observation",
            sub_grid_size=1,
            add_noise=False,
            noise_if_add_noise_false=0.2,
        )

        output_data_dir += "observation/"

        observation_ccd_data_loaded = al.load_ccd_data_from_fits(
            image_path=output_data_dir + "image.fits",
            pixel_scale=0.2,
            psf_path=output_data_dir + "psf.fits",
            noise_map_path=output_data_dir + "noise_map.fits",
            background_noise_map_path=output_data_dir + "background_noise_map.fits",
            poisson_noise_map_path=output_data_dir + "poisson_noise_map.fits",
            exposure_time_map_path=output_data_dir + "exposure_time_map.fits",
            background_sky_map_path=output_data_dir + "background_sky_map.fits",
            renormalize_psf=False,
        )

        assert (ccd_data.image == observation_ccd_data_loaded.image).all()
        assert (ccd_data.psf == observation_ccd_data_loaded.psf).all()
        assert ccd_data.noise_map == 0.2 * np.ones((11, 11))
        assert ccd_data.noise_map == observation_ccd_data_loaded.noise_map
        assert (
            ccd_data.background_noise_map
            == observation_ccd_data_loaded.background_noise_map
        )
        assert (
            ccd_data.poisson_noise_map == observation_ccd_data_loaded.poisson_noise_map
        )
        assert (
            ccd_data.exposure_time_map == observation_ccd_data_loaded.exposure_time_map
        ).all()
        assert (
            ccd_data.background_sky_map
            == observation_ccd_data_loaded.background_sky_map
        ).all()



import os
import numpy as np

# In this script, we're going to simulate a converence profile using a set of mass-profiles and perform a direct
# fit to this convergence using a model galaxy. This uses a non-linear search and the fitting process is equivalent
# to fitting an image, but it bypasses the lens modeling process (e.g. the source reconstruction and use of actual
# imaging data).

# Whilst this may sound like an odd thing to do, there are reasons why one may wish to perform a direct fit to a
# derived light or mass profile quantity, like the convergence:

# 1) If the mass-profile(s) used to generate the galaxy that is fitted and the model galaxy are different, this fit
#    will inform us of how the mismatch between profiles leads to a different estimate of the inferred mass profile
#    properties.

# 2) By bypassing the lens modeling process, we can test what results we get whilst bypass the potential systematics
#    that arise from a lens model fit (e.g due to the source reconstruction or quality of instrument).

### AUTOFIT + CONFIG SETUP ###

import autofit as af

# Setup the path to the workspace, using a relative directory name.
workspace_path = "{}/../../".format(os.path.dirname(os.path.realpath(__file__)))

# Use this path to explicitly set the config path and output path.
af.conf.instance = af.conf.Config(
    config_path=workspace_path + "config", output_path=workspace_path + "output"
)

### AUTOLENS + DATA SETUP ###

import autolens as al

# First, we'll setup the grid we use to simulate a convergence profile.
pixel_scale = 0.05
image_shape = (250, 250)
grid = al.Grid.from_shape_pixel_scale_and_sub_grid_size(
    shape=image_shape, pixel_scale=pixel_scale, sub_grid_size=8
)

# Now lets create a galaxy, using a singular isothermal sphere.
galaxy = al.Galaxy(
    redshift=0.5,
    mass=al.mass_profiles.SphericalIsothermal(centre=(0.0, 0.0), einstein_radius=1.0),
)

# Next, we'll generate its convergence profile. Note that, because we're using the galaxy convergence
# function the sub-grid of the grid that is passed to this function is used to over-sample the convergence.
# The convergence averages over this sub-grid, such that it is the shape of the image (250, 250).
convergence = galaxy.convergence_from_grid(
    galaxies=[galaxy], grid=grid, return_in_2d=True, return_binned=True
)

# Now, we'll set this convergence up as our 'galaxy-instrument', meaning that it is what we'll fit via a non-linear
# search phase. To perform a fit we need a noise-map to help define our chi-squared. Given we are fitting a direct
# lensing quantity the actual values of this noise-map arn't particularly important, so we'll just use a noise-map of
# all 0.1's
noise_map = al.ScaledSquarePixelArray(
    array=0.1 * np.ones(convergence.shape), pixel_scale=pixel_scale
)
data = al.GalaxyData(image=convergence, noise_map=noise_map, pixel_scale=pixel_scale)

# The fit will use a mask, which we setup like any other fit. Lets use a circular mask of 2.0"
def mask_function_circular(image):
    return al.Mask.circular(
        shape=image.shape, pixel_scale=image.pixel_scale, radius_arcsec=2.0
    )


# We can now use a special phase, called a 'GalaxyFitPhase', to fit this convergence with a model galaxy the
# mass-profiles of which we get to choose. We'll fit it with a singular isothermal sphere and should see we retrieve
# the input model above.
phase = al.GalaxyFitPhase(
    phase_name="galaxy_convergence_fit",
    use_convergence=True,
    galaxies=dict(
        lens=al.GalaxyModel(redshift=0.5, mass=al.mass_profiles.SphericalIsothermal)
    ),
    mask_function=mask_function_circular,
    sub_grid_size=4,
    optimizer_class=af.MultiNest,
)

phase.run(galaxy_data=[data])

# If you check your output folder, you should see that this fit has been performed and visualization specific to a
# surface-density fit is output.

# Fits to an intensity map and gravitational potential are also possible. To do so, simply change the profile quantitiy
# that is simuulated and edit the 'use_convergence' flag in the GalaxyFitPhase above to the appropriate quantity.
# You can also fit deflection angle maps - however this requires a few small changes to this script, so we've create a
# 'galaxy_fit_deflections_al.Py' example script to show you how.



from autolens.array.util import array_util

import matplotlib.pyplot as plt
import numpy as np

import os

path = "{}/".format(os.path.dirname(os.path.realpath(__file__)))

profile_name = "power_law"
# profile_name = 'ltm2_aligned'
profile_name = "ltm2_offset"
# profile_name = 'ltm2_rad_grad'
# profile_name = 'ltm2_black_hole'
# profile_name = 'uv_ltm_offset'
# profile_name = 'uv_ltm_black_hole'
# profile_name = 'uv_power_law'

file_path = (
    path
    + "/rj_lens_plots/data_original/"
    + profile_name
    + "/ML_RJLens2_All_Image_FgSub.dat"
)
output_path = (
    path + "/rj_lens_plots/data_fits/" + profile_name + "/lens_subtracted_image.fits"
)

# file_path = path+'/rj_lens_plots/data_original/'+profile_name+'/ML_RJLens_All_Recon_Src.dat'
# output_path = path+'/rj_lens_plots/data_fits/'+profile_name+'/source_model_image.fits'
#
# file_path = path+'/rj_lens_plots/data_original/'+profile_name+'/ML_RJLens_All_Recon_FgSrc.dat'
# output_path = path+'/rj_lens_plots/data_fits/'+profile_name+'/model_image.fits'
#
# file_path = path+'/rj_lens_plots/data_original/'+profile_name+'/ML_RJLens_All_Residuals.dat'
# output_path = path+'/rj_lens_plots/data_fits/'+profile_name+'/residual_map.fits'
#
# file_path = path+'/rj_lens_plots/data_original/'+profile_name+'/ML_RJLens_All_Chi_Sq_Scaled_Noise.dat'
# output_path = path+'/rj_lens_plots/data_fits/'+profile_name+'/chi_squared_map.fits'

# file_path = path+'/rj_lens_plots/data_original/'+profile_name+'/ML_RJLens_All_Scaled_Noise.dat'
# output_path = path+'/rj_lens_plots/data_fits/'+profile_name+'/noise_map.fits'

file = open(file_path, "r")

fileread = file.readlines()

image = np.zeros((422, 422))

with open(file_path, "r") as f:
    for i, lineread in enumerate(fileread):
        line = str(round(float(fileread[i][0:13]), 2))
        line += " " * (8 - len(line))
        line += str(round(float(fileread[i][13:25]), 2))
        line += " " * (16 - len(line))
        line += str(float(fileread[i][25:80])) + "\n"
        image[int(fileread[i][0:13]), int(fileread[i][13:25])] = float(
            fileread[i][25:80]
        )
    # print(int(fileread[i][0:13]), int(fileread[i][13:25]), float(fileread[i][25:80]))
    # f.write(line)

image = image.T
# image = np.fliplr(image)
image = np.flipud(image)

plt.imshow(image)
plt.show()


array_util.numpy_array_2d_to_fits(array_2d=image, file_path=output_path, overwrite=True)



import autolens as al
import os

import pytest


@pytest.fixture(name="data_plotter_path")
def make_data_plotter_setup():
    data_plotter_path = "{}/../../test_files/plotting/data/".format(
        os.path.dirname(os.path.realpath(__file__))
    )
    return data_plotter_path


def test__all_ccd_data_types_are_output(
    image_7x7,
    noise_map_7x7,
    psf_3x3,
    positions_7x7,
    mask_7x7,
    data_plotter_path,
    plot_patch,
):
    al.data_plotters.plot_image(
        image=image_7x7,
        positions=positions_7x7,
        mask=mask_7x7,
        extract_array_from_mask=True,
        zoom_around_mask=True,
        cb_tick_values=[1.0],
        cb_tick_labels=["1.0"],
        output_path=data_plotter_path,
        output_format="png",
    )

    assert data_plotter_path + "image.png" in plot_patch.paths

    al.data_plotters.plot_noise_map(
        noise_map=noise_map_7x7,
        mask=mask_7x7,
        extract_array_from_mask=True,
        zoom_around_mask=True,
        cb_tick_values=[1.0],
        cb_tick_labels=["1.0"],
        output_path=data_plotter_path,
        output_format="png",
    )

    assert data_plotter_path + "noise_map.png" in plot_patch.paths

    al.data_plotters.plot_psf(
        psf=psf_3x3,
        cb_tick_values=[1.0],
        cb_tick_labels=["1.0"],
        output_path=data_plotter_path,
        output_format="png",
    )

    assert data_plotter_path + "psf.png" in plot_patch.paths

    al.data_plotters.plot_signal_to_noise_map(
        signal_to_noise_map=image_7x7 / noise_map_7x7,
        mask=mask_7x7,
        extract_array_from_mask=True,
        zoom_around_mask=True,
        cb_tick_values=[1.0],
        cb_tick_labels=["1.0"],
        output_path=data_plotter_path,
        output_format="png",
    )

    assert data_plotter_path + "signal_to_noise_map.png" in plot_patch.paths

    al.data_plotters.plot_absolute_signal_to_noise_map(
        absolute_signal_to_noise_map=image_7x7 / noise_map_7x7,
        mask=mask_7x7,
        extract_array_from_mask=True,
        zoom_around_mask=True,
        cb_tick_values=[1.0],
        cb_tick_labels=["1.0"],
        output_path=data_plotter_path,
        output_format="png",
    )

    assert data_plotter_path + "absolute_signal_to_noise_map.png" in plot_patch.paths

    al.data_plotters.plot_potential_chi_squared_map(
        potential_chi_squared_map=image_7x7 / noise_map_7x7,
        mask=mask_7x7,
        extract_array_from_mask=True,
        zoom_around_mask=True,
        cb_tick_values=[1.0],
        cb_tick_labels=["1.0"],
        output_path=data_plotter_path,
        output_format="png",
    )

    assert data_plotter_path + "potential_chi_squared_map.png" in plot_patch.paths


def test__all_interferometer_data_types_are_output(
    visibilities_7,
    visibilities_noise_map_7,
    uv_wavelengths_7,
    primary_beam_3x3,
    data_plotter_path,
    plot_patch,
):
    al.data_plotters.plot_visibilities(
        visibilities=visibilities_7,
        visibilities_noise_map=visibilities_noise_map_7,
        cb_tick_values=[1.0],
        cb_tick_labels=["1.0"],
        output_path=data_plotter_path,
        output_format="png",
    )

    assert data_plotter_path + "visibilities.png" in plot_patch.paths

    al.data_plotters.plot_u_wavelengths(
        uv_wavelengths=uv_wavelengths_7,
        output_path=data_plotter_path,
        output_format="png",
    )

    assert data_plotter_path + "u_wavelengths.png" in plot_patch.paths

    al.data_plotters.plot_v_wavelengths(
        uv_wavelengths=uv_wavelengths_7,
        output_path=data_plotter_path,
        output_format="png",
    )

    assert data_plotter_path + "v_wavelengths.png" in plot_patch.paths

    al.data_plotters.plot_primary_beam(
        primary_beam=primary_beam_3x3,
        output_path=data_plotter_path,
        output_format="png",
    )

    assert data_plotter_path + "primary_beam.png" in plot_patch.paths



import logging
import os

import numpy as np
import pytest
import shutil
from astropy import units
from astropy.coordinates import Angle
from astropy.modeling import functional_models

import autolens as al
from autolens import exc

logger = logging.getLogger(__name__)

test_data_dir = "{}/../../test_files/array/".format(
    os.path.dirname(os.path.realpath(__file__))
)

test_positions_dir = "{}/../../test_files/positions/".format(
    os.path.dirname(os.path.realpath(__file__))
)


class TestSignalToNoise:
    def test__image_and_noise_are_values__signal_to_noise_is_ratio_of_each(self):
        array = np.array([[1.0, 2.0], [3.0, 4.0]])

        noise = np.array([[10.0, 10.0], [30.0, 4.0]])

        data = al.AbstractData(image=array, pixel_scale=1.0, psf=None, noise_map=noise)

        assert (data.signal_to_noise_map == np.array([[0.1, 0.2], [0.1, 1.0]])).all()
        assert data.signal_to_noise_max == 1.0

    def test__same_as_above__but_image_has_negative_values__replaced_with_zeros(self):
        array = np.array([[-1.0, 2.0], [3.0, -4.0]])

        noise = np.array([[10.0, 10.0], [30.0, 4.0]])

        data = al.AbstractData(image=array, pixel_scale=1.0, psf=None, noise_map=noise)

        assert (data.signal_to_noise_map == np.array([[0.0, 0.2], [0.1, 0.0]])).all()
        assert data.signal_to_noise_max == 0.2


class TestAbsoluteSignalToNoise:
    def test__image_and_noise_are_values__signal_to_noise_is_absolute_image_value_over_noise(
        self
    ):
        array = np.array([[-1.0, 2.0], [3.0, -4.0]])

        noise = np.array([[10.0, 10.0], [30.0, 4.0]])

        data = al.AbstractData(image=array, pixel_scale=1.0, psf=None, noise_map=noise)

        assert (
            data.absolute_signal_to_noise_map == np.array([[0.1, 0.2], [0.1, 1.0]])
        ).all()
        assert data.absolute_signal_to_noise_max == 1.0


class TestPotentialChiSquaredMap:
    def test__image_and_noise_are_values__signal_to_noise_is_absolute_image_value_over_noise(
        self
    ):
        array = np.array([[-1.0, 2.0], [3.0, -4.0]])

        noise = np.array([[10.0, 10.0], [30.0, 4.0]])

        data = al.AbstractData(image=array, pixel_scale=1.0, psf=None, noise_map=noise)

        assert (
            data.potential_chi_squared_map
            == np.array([[0.1 ** 2.0, 0.2 ** 2.0], [0.1 ** 2.0, 1.0 ** 2.0]])
        ).all()
        assert data.potential_chi_squared_max == 1.0


class TestAbstractNoiseMap(object):
    class TestFromWeightMap:
        def test__weight_map_no_zeros__uses_1_over_sqrt_value(self):
            weight_map = np.array([[1.0, 4.0, 16.0], [1.0, 4.0, 16.0]])

            noise_map = al.AbstractNoiseMap.from_weight_map(
                weight_map=weight_map, pixel_scale=1.0
            )

            assert (noise_map == np.array([[1.0, 0.5, 0.25], [1.0, 0.5, 0.25]])).all()
            assert noise_map.origin == (0.0, 0.0)

        def test__weight_map_no_zeros__zeros_set_to_10000000(self):
            weight_map = np.array([[1.0, 4.0, 0.0], [1.0, 4.0, 16.0]])

            noise_map = al.AbstractNoiseMap.from_weight_map(
                weight_map=weight_map, pixel_scale=1.0
            )

            assert (noise_map == np.array([[1.0, 0.5, 1.0e8], [1.0, 0.5, 0.25]])).all()
            assert noise_map.origin == (0.0, 0.0)

    class TestFromInverseAbstractNoiseMap:
        def test__inverse_noise_map_no_zeros__uses_1_over_value(self):
            inverse_noise_map = np.array([[1.0, 4.0, 16.0], [1.0, 4.0, 16.0]])

            noise_map = al.AbstractNoiseMap.from_inverse_noise_map(
                inverse_noise_map=inverse_noise_map, pixel_scale=1.0
            )

            assert (
                noise_map == np.array([[1.0, 0.25, 0.0625], [1.0, 0.25, 0.0625]])
            ).all()
            assert noise_map.origin == (0.0, 0.0)


class TestPSF(object):
    class TestConstructors(object):
        def test__init__input_psf__all_attributes_correct_including_data_inheritance(
            self
        ):
            psf = al.PSF(array=np.ones((3, 3)), pixel_scale=1.0, renormalize=False)

            assert psf.shape == (3, 3)
            assert psf.pixel_scale == 1.0
            assert (psf == np.ones((3, 3))).all()
            assert psf.origin == (0.0, 0.0)

            psf = al.PSF(array=np.ones((4, 3)), pixel_scale=1.0, renormalize=False)

            assert (psf == np.ones((4, 3))).all()
            assert psf.pixel_scale == 1.0
            assert psf.shape == (4, 3)
            assert psf.origin == (0.0, 0.0)

        def test__from_fits__input_psf_3x3__all_attributes_correct_including_data_inheritance(
            self
        ):
            psf = al.PSF.from_fits_with_scale(
                file_path=test_data_dir + "3x3_ones.fits", hdu=0, pixel_scale=1.0
            )

            assert (psf == np.ones((3, 3))).all()
            assert psf.pixel_scale == 1.0
            assert psf.origin == (0.0, 0.0)

            psf = al.PSF.from_fits_with_scale(
                file_path=test_data_dir + "4x3_ones.fits", hdu=0, pixel_scale=1.0
            )

            assert (psf == np.ones((4, 3))).all()
            assert psf.pixel_scale == 1.0
            assert psf.origin == (0.0, 0.0)

    class TestRenormalize(object):
        def test__input_is_already_normalized__no_change(self):
            psf_data = np.ones((3, 3)) / 9.0

            psf = al.PSF(array=psf_data, pixel_scale=1.0, renormalize=True)

            assert psf == pytest.approx(psf_data, 1e-3)

        def test__input_is_above_normalization_so_is_normalized(self):
            psf_data = np.ones((3, 3))

            psf = al.PSF(array=psf_data, pixel_scale=1.0, renormalize=True)

            assert psf == pytest.approx(np.ones((3, 3)) / 9.0, 1e-3)

        def test__same_as_above__renomalized_false_does_not_renormalize(self):
            psf_data = np.ones((3, 3))

            psf = al.PSF(array=psf_data, pixel_scale=1.0, renormalize=False)

            assert psf == pytest.approx(np.ones((3, 3)), 1e-3)

    class TestBinnedUp(object):
        def test__psf_is_even_x_even__rescaled_to_odd_x_odd__no_use_of_dimension_trimming(
            self
        ):
            array_2d = np.ones((6, 6))
            psf = al.PSF(array=array_2d, pixel_scale=1.0, renormalize=False)
            psf = psf.new_psf_with_rescaled_odd_dimensioned_array(
                rescale_factor=0.5, renormalize=True
            )
            assert psf.pixel_scale == 2.0
            assert psf == (1.0 / 9.0) * np.ones((3, 3))

            array_2d = np.ones((9, 9))
            psf = al.PSF(array=array_2d, pixel_scale=1.0, renormalize=False)
            psf = psf.new_psf_with_rescaled_odd_dimensioned_array(
                rescale_factor=0.333333333333333, renormalize=True
            )
            assert psf.pixel_scale == 3.0
            assert psf == (1.0 / 9.0) * np.ones((3, 3))

            array_2d = np.ones((18, 6))
            psf = al.PSF(array=array_2d, pixel_scale=1.0, renormalize=False)
            psf = psf.new_psf_with_rescaled_odd_dimensioned_array(
                rescale_factor=0.5, renormalize=True
            )
            assert psf.pixel_scale == 2.0
            assert psf == (1.0 / 27.0) * np.ones((9, 3))

            array_2d = np.ones((6, 18))
            psf = al.PSF(array=array_2d, pixel_scale=1.0, renormalize=False)
            psf = psf.new_psf_with_rescaled_odd_dimensioned_array(
                rescale_factor=0.5, renormalize=True
            )
            assert psf.pixel_scale == 2.0
            assert psf == (1.0 / 27.0) * np.ones((3, 9))

        def test__psf_is_even_x_even_after_binning_up__resized_to_odd_x_odd_with_shape_plus_one(
            self
        ):
            array_2d = np.ones((2, 2))
            psf = al.PSF(array=array_2d, pixel_scale=1.0, renormalize=False)
            psf = psf.new_psf_with_rescaled_odd_dimensioned_array(
                rescale_factor=2.0, renormalize=True
            )
            assert psf.pixel_scale == 0.4
            assert psf == (1.0 / 25.0) * np.ones((5, 5))

            array_2d = np.ones((40, 40))
            psf = al.PSF(array=array_2d, pixel_scale=1.0, renormalize=False)
            psf = psf.new_psf_with_rescaled_odd_dimensioned_array(
                rescale_factor=0.1, renormalize=True
            )
            assert psf.pixel_scale == 8.0
            assert psf == (1.0 / 25.0) * np.ones((5, 5))

            array_2d = np.ones((2, 4))
            psf = al.PSF(array=array_2d, pixel_scale=1.0, renormalize=False)
            psf = psf.new_psf_with_rescaled_odd_dimensioned_array(
                rescale_factor=2.0, renormalize=True
            )
            assert psf.pixel_scale == pytest.approx(0.4444444, 1.0e-4)
            assert psf == (1.0 / 45.0) * np.ones((5, 9))

            array_2d = np.ones((4, 2))
            psf = al.PSF(array=array_2d, pixel_scale=1.0, renormalize=False)
            psf = psf.new_psf_with_rescaled_odd_dimensioned_array(
                rescale_factor=2.0, renormalize=True
            )
            assert psf.pixel_scale == pytest.approx(0.4444444, 1.0e-4)
            assert psf == (1.0 / 45.0) * np.ones((9, 5))

        def test__psf_is_odd_and_even_after_binning_up__resized_to_odd_and_odd_with_shape_plus_one(
            self
        ):
            array_2d = np.ones((6, 4))
            psf = al.PSF(array=array_2d, pixel_scale=1.0, renormalize=False)
            psf = psf.new_psf_with_rescaled_odd_dimensioned_array(
                rescale_factor=0.5, renormalize=True
            )
            assert psf.pixel_scale == pytest.approx(2.0, 1.0e-4)
            assert psf == (1.0 / 9.0) * np.ones((3, 3))

            array_2d = np.ones((9, 12))
            psf = al.PSF(array=array_2d, pixel_scale=1.0, renormalize=False)
            psf = psf.new_psf_with_rescaled_odd_dimensioned_array(
                rescale_factor=0.33333333333, renormalize=True
            )
            assert psf.pixel_scale == pytest.approx(3.0, 1.0e-4)
            assert psf == (1.0 / 15.0) * np.ones((3, 5))

            array_2d = np.ones((4, 6))
            psf = al.PSF(array=array_2d, pixel_scale=1.0, renormalize=False)
            psf = psf.new_psf_with_rescaled_odd_dimensioned_array(
                rescale_factor=0.5, renormalize=True
            )
            assert psf.pixel_scale == pytest.approx(2.0, 1.0e-4)
            assert psf == (1.0 / 9.0) * np.ones((3, 3))

            array_2d = np.ones((12, 9))
            psf = al.PSF(array=array_2d, pixel_scale=1.0, renormalize=False)
            psf = psf.new_psf_with_rescaled_odd_dimensioned_array(
                rescale_factor=0.33333333333, renormalize=True
            )
            assert psf.pixel_scale == pytest.approx(3.0, 1.0e-4)
            assert psf == (1.0 / 15.0) * np.ones((5, 3))

    class TestNewRenormalizedPsf(object):
        def test__input_is_already_normalized__no_change(self):
            psf_data = np.ones((3, 3)) / 9.0

            psf = al.PSF(array=psf_data, pixel_scale=1.0, renormalize=False)

            psf_new = psf.new_psf_with_renormalized_array()

            assert psf_new == pytest.approx(psf_data, 1e-3)

        def test__input_is_above_normalization_so_is_normalized(self):
            psf_data = np.ones((3, 3))

            psf = al.PSF(array=psf_data, pixel_scale=1.0, renormalize=False)

            psf_new = psf.new_psf_with_renormalized_array()

            assert psf_new == pytest.approx(np.ones((3, 3)) / 9.0, 1e-3)

    class TestConvolve(object):
        def test__kernel_is_not_odd_x_odd__raises_error(self):
            kernel = np.array([[0.0, 1.0], [1.0, 2.0]])

            psf = al.PSF(array=kernel, pixel_scale=1.0)

            with pytest.raises(exc.ConvolutionException):
                psf.convolve(np.ones((5, 5)))

        def test__image_is_3x3_central_value_of_one__kernel_is_cross__blurred_image_becomes_cross(
            self
        ):
            image = np.array([[0.0, 0.0, 0.0], [0.0, 1.0, 0.0], [0.0, 0.0, 0.0]])

            kernel = np.array([[0.0, 1.0, 0.0], [1.0, 2.0, 1.0], [0.0, 1.0, 0.0]])

            psf = al.PSF(array=kernel, pixel_scale=1.0)

            blurred_image = psf.convolve(image)

            assert (blurred_image == kernel).all()

        def test__image_is_4x4_central_value_of_one__kernel_is_cross__blurred_image_becomes_cross(
            self
        ):
            image = np.array(
                [
                    [0.0, 0.0, 0.0, 0.0],
                    [0.0, 1.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0],
                ]
            )

            kernel = np.array([[0.0, 1.0, 0.0], [1.0, 2.0, 1.0], [0.0, 1.0, 0.0]])

            psf = al.PSF(array=kernel, pixel_scale=1.0)

            blurred_image = psf.convolve(image)

            assert (
                blurred_image
                == np.array(
                    [
                        [0.0, 1.0, 0.0, 0.0],
                        [1.0, 2.0, 1.0, 0.0],
                        [0.0, 1.0, 0.0, 0.0],
                        [0.0, 0.0, 0.0, 0.0],
                    ]
                )
            ).all()

        def test__image_is_4x3_central_value_of_one__kernel_is_cross__blurred_image_becomes_cross(
            self
        ):
            image = np.array(
                [[0.0, 0.0, 0.0], [0.0, 1.0, 0.0], [0.0, 0.0, 0.0], [0.0, 0.0, 0.0]]
            )

            kernel = np.array([[0.0, 1.0, 0.0], [1.0, 2.0, 1.0], [0.0, 1.0, 0.0]])

            psf = al.PSF(array=kernel, pixel_scale=1.0)

            blurred_image = psf.convolve(image)

            assert (
                blurred_image
                == np.array(
                    [[0.0, 1.0, 0.0], [1.0, 2.0, 1.0], [0.0, 1.0, 0.0], [0.0, 0.0, 0.0]]
                )
            ).all()

        def test__image_is_3x4_central_value_of_one__kernel_is_cross__blurred_image_becomes_cross(
            self
        ):
            image = np.array(
                [[0.0, 0.0, 0.0, 0.0], [0.0, 1.0, 0.0, 0.0], [0.0, 0.0, 0.0, 0.0]]
            )

            kernel = np.array([[0.0, 1.0, 0.0], [1.0, 2.0, 1.0], [0.0, 1.0, 0.0]])

            psf = al.PSF(array=kernel, pixel_scale=1.0)

            blurred_image = psf.convolve(image)

            assert (
                blurred_image
                == np.array(
                    [[0.0, 1.0, 0.0, 0.0], [1.0, 2.0, 1.0, 0.0], [0.0, 1.0, 0.0, 0.0]]
                )
            ).all()

        def test__image_is_4x4_has_two_central_values__kernel_is_asymmetric__blurred_image_follows_convolution(
            self
        ):
            image = np.array(
                [
                    [0.0, 0.0, 0.0, 0.0],
                    [0.0, 1.0, 0.0, 0.0],
                    [0.0, 0.0, 1.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0],
                ]
            )

            kernel = np.array([[1.0, 1.0, 1.0], [2.0, 2.0, 1.0], [1.0, 3.0, 3.0]])

            psf = al.PSF(array=kernel, pixel_scale=1.0)

            blurred_image = psf.convolve(image)

            assert (
                blurred_image
                == np.array(
                    [
                        [1.0, 1.0, 1.0, 0.0],
                        [2.0, 3.0, 2.0, 1.0],
                        [1.0, 5.0, 5.0, 1.0],
                        [0.0, 1.0, 3.0, 3.0],
                    ]
                )
            ).all()

        def test__image_is_4x4_values_are_on_edge__kernel_is_asymmetric__blurring_does_not_account_for_edge_effects(
            self
        ):
            image = np.array(
                [
                    [0.0, 0.0, 0.0, 0.0],
                    [1.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 1.0],
                    [0.0, 0.0, 0.0, 0.0],
                ]
            )

            kernel = np.array([[1.0, 1.0, 1.0], [2.0, 2.0, 1.0], [1.0, 3.0, 3.0]])

            psf = al.PSF(array=kernel, pixel_scale=1.0)

            blurred_image = psf.convolve(image)

            assert (
                blurred_image
                == np.array(
                    [
                        [1.0, 1.0, 0.0, 0.0],
                        [2.0, 1.0, 1.0, 1.0],
                        [3.0, 3.0, 2.0, 2.0],
                        [0.0, 0.0, 1.0, 3.0],
                    ]
                )
            ).all()

        def test__image_is_4x4_values_are_on_corner__kernel_is_asymmetric__blurring_does_not_account_for_edge_effects(
            self
        ):
            image = np.array(
                [
                    [1.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 1.0],
                ]
            )

            kernel = np.array([[1.0, 1.0, 1.0], [2.0, 2.0, 1.0], [1.0, 3.0, 3.0]])

            psf = al.PSF(array=kernel, pixel_scale=1.0)

            blurred_image = psf.convolve(image)

            assert (
                blurred_image
                == np.array(
                    [
                        [2.0, 1.0, 0.0, 0.0],
                        [3.0, 3.0, 0.0, 0.0],
                        [0.0, 0.0, 1.0, 1.0],
                        [0.0, 0.0, 2.0, 2.0],
                    ]
                )
            ).all()

    class TestFromKernelNoBlurring(object):
        def test__correct_kernel(self):
            psf = al.PSF.from_no_blurring_kernel(pixel_scale=1.0)

            assert (
                psf == np.array([[0.0, 0.0, 0.0], [0.0, 1.0, 0.0], [0.0, 0.0, 0.0]])
            ).all()
            assert psf.pixel_scale == 1.0

            psf = al.PSF.from_no_blurring_kernel(pixel_scale=2.0)

            assert (
                psf == np.array([[0.0, 0.0, 0.0], [0.0, 1.0, 0.0], [0.0, 0.0, 0.0]])
            ).all()
            assert psf.pixel_scale == 2.0

    class TestFromGaussian(object):
        def test__identical_to_gaussian_light_profile(self):
            grid = al.Grid.from_shape_pixel_scale_and_sub_grid_size(
                shape=(3, 3), pixel_scale=1.0, sub_grid_size=1
            )

            gaussian = al.light_profiles.EllipticalGaussian(
                centre=(0.1, 0.1), axis_ratio=0.9, phi=45.0, intensity=1.0, sigma=1.0
            )
            profile_gaussian = gaussian.profile_image_from_grid(
                grid=grid, return_in_2d=True, return_binned=True
            )

            profile_psf = al.PSF(
                array=profile_gaussian, pixel_scale=1.0, renormalize=True
            )

            imaging_psf = al.PSF.from_gaussian(
                shape=(3, 3),
                pixel_scale=1.0,
                centre=(0.1, 0.1),
                axis_ratio=0.9,
                phi=45.0,
                sigma=1.0,
            )

            assert profile_psf == pytest.approx(imaging_psf, 1e-4)

    class TestFromAlmaGaussian(object):
        def test__identical_to_astropy_gaussian_model__circular_no_rotation(self):
            pixel_scale = 0.1

            x_stddev = (
                2.0e-5
                * (units.deg).to(units.arcsec)
                / pixel_scale
                / (2.0 * np.sqrt(2.0 * np.log(2.0)))
            )
            y_stddev = (
                2.0e-5
                * (units.deg).to(units.arcsec)
                / pixel_scale
                / (2.0 * np.sqrt(2.0 * np.log(2.0)))
            )

            gaussian_astropy = functional_models.Gaussian2D(
                amplitude=1.0,
                x_mean=2.0,
                y_mean=2.0,
                x_stddev=x_stddev,
                y_stddev=y_stddev,
                theta=0.0,
            )

            shape = (5, 5)
            y, x = np.mgrid[0 : shape[1], 0 : shape[0]]
            psf_astropy = gaussian_astropy(x, y)
            psf_astropy /= np.sum(psf_astropy)

            psf = al.PSF.from_as_gaussian_via_alma_fits_header_parameters(
                shape=shape,
                pixel_scale=pixel_scale,
                y_stddev=2.0e-5,
                x_stddev=2.0e-5,
                theta=0.0,
            )

            assert psf_astropy == pytest.approx(psf, 1e-4)

        def test__identical_to_astropy_gaussian_model__circular_no_rotation_different_pixel_scale(
            self
        ):
            pixel_scale = 0.02

            x_stddev = (
                2.0e-5
                * (units.deg).to(units.arcsec)
                / pixel_scale
                / (2.0 * np.sqrt(2.0 * np.log(2.0)))
            )
            y_stddev = (
                2.0e-5
                * (units.deg).to(units.arcsec)
                / pixel_scale
                / (2.0 * np.sqrt(2.0 * np.log(2.0)))
            )

            gaussian_astropy = functional_models.Gaussian2D(
                amplitude=1.0,
                x_mean=2.0,
                y_mean=2.0,
                x_stddev=x_stddev,
                y_stddev=y_stddev,
                theta=0.0,
            )

            shape = (5, 5)
            y, x = np.mgrid[0 : shape[1], 0 : shape[0]]
            psf_astropy = gaussian_astropy(x, y)
            psf_astropy /= np.sum(psf_astropy)

            psf = al.PSF.from_as_gaussian_via_alma_fits_header_parameters(
                shape=shape,
                pixel_scale=pixel_scale,
                y_stddev=2.0e-5,
                x_stddev=2.0e-5,
                theta=0.0,
            )

            assert psf_astropy == pytest.approx(psf, 1e-4)

        def test__identical_to_astropy_gaussian_model__include_ellipticity_from_x_and_y_stddev(
            self
        ):
            pixel_scale = 0.1

            x_stddev = (
                1.0e-5
                * (units.deg).to(units.arcsec)
                / pixel_scale
                / (2.0 * np.sqrt(2.0 * np.log(2.0)))
            )
            y_stddev = (
                2.0e-5
                * (units.deg).to(units.arcsec)
                / pixel_scale
                / (2.0 * np.sqrt(2.0 * np.log(2.0)))
            )

            theta_deg = 0.0
            theta = Angle(theta_deg, "deg").radian

            gaussian_astropy = functional_models.Gaussian2D(
                amplitude=1.0,
                x_mean=2.0,
                y_mean=2.0,
                x_stddev=x_stddev,
                y_stddev=y_stddev,
                theta=theta,
            )

            shape = (5, 5)
            y, x = np.mgrid[0 : shape[1], 0 : shape[0]]
            psf_astropy = gaussian_astropy(x, y)
            psf_astropy /= np.sum(psf_astropy)

            psf = al.PSF.from_as_gaussian_via_alma_fits_header_parameters(
                shape=shape,
                pixel_scale=pixel_scale,
                y_stddev=2.0e-5,
                x_stddev=1.0e-5,
                theta=theta_deg,
            )

            assert psf_astropy == pytest.approx(psf, 1e-4)

        def test__identical_to_astropy_gaussian_model__include_different_ellipticity_from_x_and_y_stddev(
            self
        ):
            pixel_scale = 0.1

            x_stddev = (
                3.0e-5
                * (units.deg).to(units.arcsec)
                / pixel_scale
                / (2.0 * np.sqrt(2.0 * np.log(2.0)))
            )
            y_stddev = (
                2.0e-5
                * (units.deg).to(units.arcsec)
                / pixel_scale
                / (2.0 * np.sqrt(2.0 * np.log(2.0)))
            )

            theta_deg = 0.0
            theta = Angle(theta_deg, "deg").radian

            gaussian_astropy = functional_models.Gaussian2D(
                amplitude=1.0,
                x_mean=2.0,
                y_mean=2.0,
                x_stddev=x_stddev,
                y_stddev=y_stddev,
                theta=theta,
            )

            shape = (5, 5)
            y, x = np.mgrid[0 : shape[1], 0 : shape[0]]
            psf_astropy = gaussian_astropy(x, y)
            psf_astropy /= np.sum(psf_astropy)

            psf = al.PSF.from_as_gaussian_via_alma_fits_header_parameters(
                shape=shape,
                pixel_scale=pixel_scale,
                y_stddev=2.0e-5,
                x_stddev=3.0e-5,
                theta=theta_deg,
            )

            assert psf_astropy == pytest.approx(psf, 1e-4)

        def test__identical_to_astropy_gaussian_model__include_rotation_angle_30(self):
            pixel_scale = 0.1

            x_stddev = (
                1.0e-5
                * (units.deg).to(units.arcsec)
                / pixel_scale
                / (2.0 * np.sqrt(2.0 * np.log(2.0)))
            )
            y_stddev = (
                2.0e-5
                * (units.deg).to(units.arcsec)
                / pixel_scale
                / (2.0 * np.sqrt(2.0 * np.log(2.0)))
            )

            theta_deg = 30.0
            theta = Angle(theta_deg, "deg").radian

            gaussian_astropy = functional_models.Gaussian2D(
                amplitude=1.0,
                x_mean=1.0,
                y_mean=1.0,
                x_stddev=x_stddev,
                y_stddev=y_stddev,
                theta=theta,
            )

            shape = (3, 3)
            y, x = np.mgrid[0 : shape[1], 0 : shape[0]]
            psf_astropy = gaussian_astropy(x, y)
            psf_astropy /= np.sum(psf_astropy)

            psf = al.PSF.from_as_gaussian_via_alma_fits_header_parameters(
                shape=shape,
                pixel_scale=pixel_scale,
                y_stddev=2.0e-5,
                x_stddev=1.0e-5,
                theta=theta_deg,
            )

            assert psf_astropy == pytest.approx(psf, 1e-4)

        def test__identical_to_astropy_gaussian_model__include_rotation_angle_230(self):
            pixel_scale = 0.1

            x_stddev = (
                1.0e-5
                * (units.deg).to(units.arcsec)
                / pixel_scale
                / (2.0 * np.sqrt(2.0 * np.log(2.0)))
            )
            y_stddev = (
                2.0e-5
                * (units.deg).to(units.arcsec)
                / pixel_scale
                / (2.0 * np.sqrt(2.0 * np.log(2.0)))
            )

            theta_deg = 230.0
            theta = Angle(theta_deg, "deg").radian

            gaussian_astropy = functional_models.Gaussian2D(
                amplitude=1.0,
                x_mean=1.0,
                y_mean=1.0,
                x_stddev=x_stddev,
                y_stddev=y_stddev,
                theta=theta,
            )

            shape = (3, 3)
            y, x = np.mgrid[0 : shape[1], 0 : shape[0]]
            psf_astropy = gaussian_astropy(x, y)
            psf_astropy /= np.sum(psf_astropy)

            psf = al.PSF.from_as_gaussian_via_alma_fits_header_parameters(
                shape=shape,
                pixel_scale=pixel_scale,
                y_stddev=2.0e-5,
                x_stddev=1.0e-5,
                theta=theta_deg,
            )

            assert psf_astropy == pytest.approx(psf, 1e-4)


class TestExposureTimeMap(object):
    class TestFromExposureTimeAndBackgroundNoiseMap:
        def test__from_background_noise_map__covnerts_to_exposure_times(self):
            background_noise_map = np.array([[1.0, 4.0, 8.0], [1.0, 4.0, 8.0]])

            exposure_time_map = al.ExposureTimeMap.from_exposure_time_and_inverse_noise_map(
                pixel_scale=0.1,
                exposure_time=1.0,
                inverse_noise_map=background_noise_map,
            )

            assert (
                exposure_time_map == np.array([[0.125, 0.5, 1.0], [0.125, 0.5, 1.0]])
            ).all()
            assert exposure_time_map.origin == (0.0, 0.0)

            exposure_time_map = al.ExposureTimeMap.from_exposure_time_and_inverse_noise_map(
                pixel_scale=0.1,
                exposure_time=3.0,
                inverse_noise_map=background_noise_map,
            )

            assert (
                exposure_time_map == np.array([[0.375, 1.5, 3.0], [0.375, 1.5, 3.0]])
            ).all()
            assert exposure_time_map.origin == (0.0, 0.0)


class TestDataFromFits(object):
    def test__psf_renormalized_true__renormalized_psf(self):
        psf = al.abstract_data.load_psf(
            pixel_scale=0.1,
            psf_path=test_data_dir + "3x3_twos.fits",
            psf_hdu=0,
            renormalize=True,
        )

        assert psf == pytest.approx((1.0 / 9.0) * np.ones((3, 3)), 1e-2)


class TestPositionsToFile(object):
    def test__load_positions__retains_list_structure(self):
        positions = al.abstract_data.load_positions(
            positions_path=test_positions_dir + "positions_test.dat"
        )

        assert positions == [
            [[1.0, 1.0], [2.0, 2.0]],
            [[3.0, 3.0], [4.0, 4.0], [5.0, 6.0]],
        ]

    def test__output_positions(self):
        positions = [[[4.0, 4.0], [5.0, 5.0]], [[6.0, 6.0], [7.0, 7.0], [8.0, 8.0]]]

        output_data_dir = "{}/../test_files/positions/output_test/".format(
            os.path.dirname(os.path.realpath(__file__))
        )
        if os.path.exists(output_data_dir):
            shutil.rmtree(output_data_dir)

        os.makedirs(output_data_dir)

        al.abstract_data.output_positions(
            positions=positions, positions_path=output_data_dir + "positions_test.dat"
        )

        positions = al.abstract_data.load_positions(
            positions_path=output_data_dir + "positions_test.dat"
        )

        assert positions == [
            [[4.0, 4.0], [5.0, 5.0]],
            [[6.0, 6.0], [7.0, 7.0], [8.0, 8.0]],
        ]



import autofit as af
import autolens as al

# In this pipeline, we show how custom positions (generated using the tools/positions_maker.py file) can be input and
# used by a pipeline. Both phases will be input with a 'positions_threshold', which requires that a set of positions
# drawn on the image in the image plane trace within this threshold for a given mass model. If they don't, the mass
# model is instantly discarded and resampled. This provides two benefits:

# 1) PyAutoLens does not waste computing time on mass models which will clearly give a poor fit to the data.

# 2) By discarding these mass models, non-linear parameter space will be less complex and thus easier to sample.

# To use positions, the positions must be passed to the pipeline's 'run' function. See runners/runner_positions.py
# for an example of this.

# We use a simple two phase pipeline:

# Phase 1:

# Description: Fits the lens and source model using positions to resample mass models.
# Lens Mass: EllipitcalIsothermal
# Source Light: EllipticalSersic
# Previous Pipelines: None
# Prior Passing: None
# Notes: Use positions to resample mass models.

# Phase 1:

# Description: Fits the lens and source model without using position thresholding.
# Lens Mass: EllipitcalIsothermal
# Source Light: EllipticalSersic
# Previous Pipelines: None
# Prior Passing: Lens mass (variable -> phase 1), source light (variable -> phase 1)
# Notes: No position thresholding.


def make_pipeline(phase_folders=None, positions_threshold=None):

    ### SETUP PIPELINE AND PHASE NAMES, TAGS AND PATHS ###

    # We setup the pipeline name using the tagging module. In this case, the pipeline name is not given a tag and
    # will be the string specified below However, its good practise to use the 'tag.' function below, incase
    # a pipeline does use customized tag names.

    pipeline_name = "pipeline_feature__position_thresholding"

    pipeline_tag = al.pipeline_tagging.pipeline_tag_from_pipeline_settings()

    # This function uses the phase folders and pipeline name to set up the output directory structure,
    # e.g. 'autolens_workspace/output/phase_folder_1/phase_folder_2/pipeline_name/phase_name/'

    phase_folders.append(pipeline_name)
    phase_folders.append(pipeline_tag)

    # A settings tag is automatically added to the phase path, making it clear the position threshold value used.
    # The positions_threshold_tag and phase name are shown for 3 example inner mask radii values:

    # positions_threshold=0.2 -> phase_path='phase_name/settings_pos_0.20'
    # positions_threshold=0.25, -> phase_path='phase_name/settings_pos_0.25'

    # If the positions_threshold is None, the tag is an empty string, thus not changing the phase name:

    # - positions_threshold=None, positions_threshold_tag='', phase_name=phase_name/settings

    ### PHASE 1 ###

    # In phase 1, we will:

    # 1) Input the positions_threshold, such that if the value is not None it is used to resample mass models.

    def mask_function(image):
        return al.Mask.circular(
            shape=image.shape, pixel_scale=image.pixel_scale, radius_arcsec=2.5
        )

    phase1 = al.PhaseImaging(
        phase_name="phase_1__use_positions",
        phase_folders=phase_folders,
        galaxies=dict(
            lens=al.GalaxyModel(
                redshift=0.5, mass=al.mass_profiles.EllipticalIsothermal
            ),
            source=al.GalaxyModel(
                redshift=1.0, light=al.light_profiles.EllipticalSersic
            ),
        ),
        mask_function=mask_function,
        positions_threshold=0.3,
        optimizer_class=af.MultiNest,
    )

    phase1.optimizer.const_efficiency_mode = True
    phase1.optimizer.n_live_points = 30
    phase1.optimizer.sampling_efficiency = 0.3

    ### PHASE 2 ###

    # In phase 2, we will:

    # 1) Initialize the priors on the lens galaxy's mass and source galaxy's light by linking them to those inferred
    #    in phase 1.

    # 2) Not specify a positions_threshold, such that is defaults to None in the phase and is not used to resample
    #        mass models.

    phase2 = al.PhaseImaging(
        phase_name="phase_2__no_positions",
        phase_folders=phase_folders,
        galaxies=dict(
            lens=al.GalaxyModel(
                redshift=0.5, mass=phase1.result.variable.galaxies.lens.mass
            ),
            source=al.GalaxyModel(
                redshift=1.0, light=phase1.result.variable.galaxies.source.light
            ),
        ),
        mask_function=mask_function,
        positions_threshold=positions_threshold,
        optimizer_class=af.MultiNest,
    )

    phase2.optimizer.const_efficiency_mode = True
    phase2.optimizer.n_live_points = 50
    phase2.optimizer.sampling_efficiency = 0.2

    return al.PipelineImaging(pipeline_name, phase1, phase2)



from autolens import text_util
from autolens.array import grids
from autolens.array.grids import (
    Grid,
    BinnedGrid,
    PixelizationGrid,
    SparseToGrid,
    Interpolator,
)
from autolens.array.mapping_util import array_mapping_util
from autolens.array.mapping_util import grid_mapping_util
from autolens.array.mapping_util import mask_mapping_util
from autolens.array.mapping_util import mask_mapping_util, sparse_mapping_util
from autolens.array.mapping_util import sparse_mapping_util
from autolens.array.mask import Mask
from autolens.array.mask import load_mask_from_fits, output_mask_to_fits
from autolens.array.scaled_array import (
    ArrayGeometry,
    RectangularArrayGeometry,
    Array,
    ScaledArray,
    ScaledSquarePixelArray,
    ScaledRectangularPixelArray,
)
from autolens.array.util import array_util, grid_util, binning_util
from autolens.array.util import binning_util, grid_util, mask_util
from autolens.array.util import grid_util, mask_util
from autolens.data.convolution import Convolver
from autolens.data.fourier_transform import Transformer
from autolens.data.instrument import abstract_data
from autolens.data.instrument import interferometer
from autolens.data.instrument.abstract_data import (
    PSF,
    AbstractData,
    AbstractNoiseMap,
    ExposureTimeMap,
    load_image,
    load_psf,
    load_exposure_time_map,
    load_positions,
    output_positions,
)
from autolens.data.instrument.ccd import (
    CCDData,
    NoiseMap,
    PoissonNoiseMap,
    SimulatedCCDData,
    generate_poisson_noise,
    load_ccd_data_from_fits,
    load_noise_map,
    load_ccd_data_from_fits,
    output_ccd_data_to_fits,
)
from autolens.data.instrument.interferometer import (
    InterferometerData,
    PrimaryBeam,
    SimulatedInterferometerData,
    load_interferometer_data_from_fits,
    output_interferometer_data_to_fits,
    gaussian_noise_map_from_shape_and_sigma,
)
from autolens.data.plotters import ccd_plotters
from autolens.data.plotters import data_plotters
from autolens.data.plotters import interferometer_plotters
from autolens.dimensions import (
    DimensionsProfile,
    Length,
    Luminosity,
    Mass,
    MassOverLuminosity,
    MassOverLength2,
    MassOverLength3,
    Position,
    convert_units_to_input_units,
)
from autolens.lens import ray_tracing, lens_fit
from autolens.lens.lens_data import LensData
from autolens.lens.lens_fit import (
    LensDataFit,
    LensTracerFit,
    InversionFit,
    LensInversionFit,
    LensProfileInversionFit,
    LensPositionFit,
    LensProfileFit,
)
from autolens.lens.plane import Plane, PlanePositions, PlaneImage
from autolens.lens.plotters import lens_fit_plotters
from autolens.lens.plotters import lens_plotter_util
from autolens.lens.plotters import plane_plotters
from autolens.lens.plotters import ray_tracing_plotters
from autolens.lens.ray_tracing import Tracer
from autolens.lens.util import lens_fit_util
from autolens.lens.util import lens_util
from autolens.model import cosmology_util
from autolens.model.galaxy.galaxy import Galaxy
from autolens.model.galaxy.galaxy import HyperGalaxy
from autolens.model.galaxy.galaxy import Redshift
from autolens.model.galaxy.galaxy_data import GalaxyData, GalaxyFitData
from autolens.model.galaxy.galaxy_fit import GalaxyFit
from autolens.model.galaxy.galaxy_model import GalaxyModel
from autolens.model.galaxy.plotters import galaxy_fit_plotters
from autolens.model.galaxy.plotters import galaxy_plotters
from autolens.model.hyper.hyper_data import HyperImageSky, HyperBackgroundNoise
from autolens.model.inversion.inversions import Inversion
from autolens.model.inversion.mappers import Mapper, RectangularMapper, VoronoiMapper
from autolens.model.inversion import pixelizations
from autolens.model.inversion.plotters import inversion_plotters, mapper_plotters
from autolens.model.inversion import regularization
from autolens.model.inversion.util import inversion_util
from autolens.model.inversion.util import mapper_util
from autolens.model.inversion.util import pixelization_util
from autolens.model.inversion.util import regularization_util
from autolens.model.profiles import geometry_profiles
from autolens.model.profiles import light_and_mass_profiles
from autolens.model.profiles import light_profiles
from autolens.model.profiles import mass_profiles
from autolens.model.profiles.plotters import profile_plotters
from autolens.pipeline import phase_tagging
from autolens.pipeline import pipeline_tagging
from autolens.pipeline.phase import GalaxyFitPhase, Phase
from autolens.pipeline.phase import phase
from autolens.pipeline.phase.phase import AbstractPhase
from autolens.pipeline.phase.phase_extensions import CombinedHyperPhase
from autolens.pipeline.phase.phase_extensions import HyperGalaxyPhase
from autolens.pipeline.phase.phase_extensions.hyper_galaxy_phase import HyperGalaxyPhase
from autolens.pipeline.phase.phase_extensions.hyper_phase import HyperPhase
from autolens.pipeline.phase.phase_extensions.inversion_phase import (
    InversionBackgroundBothPhase,
)
from autolens.pipeline.phase.phase_extensions.inversion_phase import (
    InversionBackgroundNoisePhase,
)
from autolens.pipeline.phase.phase_extensions.inversion_phase import (
    InversionBackgroundSkyPhase,
)
from autolens.pipeline.phase.phase_extensions.inversion_phase import InversionPhase
from autolens.pipeline.phase.phase_extensions.inversion_phase import (
    VariableFixingHyperPhase,
)
from autolens.pipeline.phase.phase_imaging import PhaseImaging
from autolens.pipeline.phase.phase_positions import PhasePositions
from autolens.pipeline.pipeline import (
    PipelineSettings,
    PipelineSettingsHyper,
    PipelineImaging,
    PipelinePositions,
)
from autolens.pipeline.plotters import hyper_plotters
from autolens.pipeline.plotters import phase_plotters
from autolens.plotters import array_plotters, grid_plotters, plotter_util

__version__ = '0.30.4'



from autolens import exc
from autolens.array import grids, scaled_array
from autolens.array import mask as msk

from autolens.array.grids import reshape_array


class GalaxyData(object):
    def __init__(self, image, noise_map, pixel_scale):
        """ A galaxy-fit instrument is a collection of fit instrument components which are used to fit a galaxy to another galaxy. \
        This is where a component of a galaxy's light profiles (e.g. image) or mass profiles (e.g. convergence \
        , potential or deflection angles) are fitted to one another.

        This is primarily performed for automatic prior linking, as a means to efficiently link the priors of a galaxy \
        using one inferred parametrization of light or mass profiles to a new galaxy with a different parametrization \
        of light or mass profiles.

        This omits a number of the fit instrument components typically used when fitting an image (e.g. the observed image, PSF, \
        exposure time map), but still has a number of the other components (e.g. an effective noise_map-map, grid_stacks).

        Parameters
        ----------
        image : scaled_array.ScaledSquarePixelArray
            An image of the quantity of the galaxy that is being fitted (e.g. its image, convergence, etc.).
        noise_map : scaled_array.ScaledSquarePixelArray
            The noise_map-map used for computing the likelihood of each fit. This can be chosen arbritarily.
        """
        self.image = image
        self.noise_map = noise_map
        self.pixel_scale = pixel_scale


class GalaxyFitData(object):
    def __init__(
        self,
        galaxy_data,
        mask,
        sub_grid_size=2,
        pixel_scale_interpolation_grid=None,
        use_image=False,
        use_convergence=False,
        use_potential=False,
        use_deflections_y=False,
        use_deflections_x=False,
    ):
        """ A galaxy-fit instrument is a collection of fit instrument components which are used to fit a galaxy to another galaxy. \
        This is where a component of a galaxy's light profiles (e.g. image) or mass profiles (e.g. surface \
        density, potential or deflection angles) are fitted to one another.

        This is primarily performed for automatic prior linking, as a means to efficiently link the priors of a galaxy \
        using one inferred parametrization of light or mass profiles to a new galaxy with a different parametrization \
        of light or mass profiles.

        This omits a number of the fit instrument components typically used when fitting an image (e.g. the observed image, PSF, \
        exposure time map), but still has a number of the other components (e.g. an effective noise_map-map, grid_stacks).

        Parameters
        ----------
        galaxy_data : GalaxyData
            The collection of instrument about the galaxy (image of its profile map, noise-map, etc.) that is fitted.
        mask: msk.Mask
            The 2D masks that is applied to image fit instrument.
        sub_grid_size : int
            The size of the sub-grid used for computing the SubGrid (see ccd.masks.SubGrid).

        Attributes
        ----------
        noise_map_1d : ndarray
            The masked 1D array of the noise_map-map
        grid_stacks : ccd.masks.GridStack
            Grids of (y,x) Cartesian coordinates which map over the masked 1D fit instrument array's pixels (includes an \
            grid, sub-grid, etc.)
        """
        self.unmasked_image = galaxy_data.image
        self.pixel_scale = galaxy_data.pixel_scale
        self.unmasked_noise_map = galaxy_data.noise_map

        self.image_1d = mask.array_1d_from_array_2d(array_2d=self.unmasked_image)
        self.noise_map_1d = mask.array_1d_from_array_2d(
            array_2d=self.unmasked_noise_map
        )
        self.signal_to_noise_map_1d = self.image_1d / self.noise_map_1d
        self.mask_1d = mask.array_1d_from_array_2d(array_2d=mask)

        self.sub_grid_size = sub_grid_size

        self.grid = grids.Grid.from_mask_and_sub_grid_size(
            mask=mask, sub_grid_size=sub_grid_size
        )

        self.pixel_scale_interpolation_grid = pixel_scale_interpolation_grid

        if pixel_scale_interpolation_grid is not None:

            self.grid = self.grid.new_grid_with_interpolator(
                pixel_scale_interpolation_grid=pixel_scale_interpolation_grid
            )

        self.mask_2d = mask

        if all(
            not element
            for element in [
                use_image,
                use_convergence,
                use_potential,
                use_deflections_y,
                use_deflections_x,
            ]
        ):
            raise exc.GalaxyException(
                "The galaxy fit instrument has not been supplied with a use_ method."
            )

        if (
            sum(
                [
                    use_image,
                    use_convergence,
                    use_potential,
                    use_deflections_y,
                    use_deflections_x,
                ]
            )
            > 1
        ):
            raise exc.GalaxyException(
                "The galaxy fit instrument has not been supplied with multiple use_ methods, only supply "
                "one."
            )

        self.use_image = use_image
        self.use_convergence = use_convergence
        self.use_potential = use_potential
        self.use_deflections_y = use_deflections_y
        self.use_deflections_x = use_deflections_x

    def __array_finalize__(self, obj):
        super(GalaxyFitData, self).__array_finalize__(obj)
        if isinstance(obj, GalaxyFitData):
            self.unmasked_image = obj.unmasked_image
            self.pixel_scale = obj.pixel_scale
            self.mask_2d = obj.mask_2d
            self.unmasked_noise_map = obj.unmasked_noise_map
            self.image_1d = obj.image_1d
            self.noise_map_1d = obj.noise_map_1d
            self.mask_1d = obj.mask_1d
            self.sub_grid_size = obj.sub_grid_size
            self.pixel_scale_interpolation_grid = obj.pixel_scale_interpolation_grid
            self.grid = obj.grid
            self.use_image = obj.use_image
            self.use_convergence = obj.use_convergence
            self.use_potential = obj.use_potential
            self.use_deflections_y = obj.use_deflections_y
            self.use_deflections_x = obj.use_deflections_x

    def map_to_scaled_array(self, array_1d):
        return self.grid.scaled_array_2d_from_array_1d(array_1d=array_1d)

    def profile_quantity_from_galaxies(self, galaxies):

        if self.use_image:
            return sum(
                map(
                    lambda g: g.profile_image_from_grid(
                        grid=self.grid, return_in_2d=False, return_binned=True
                    ),
                    galaxies,
                )
            )
        elif self.use_convergence:
            return sum(
                map(
                    lambda g: g.convergence_from_grid(
                        grid=self.grid, return_in_2d=False, return_binned=True
                    ),
                    galaxies,
                )
            )
        elif self.use_potential:
            return sum(
                map(
                    lambda g: g.potential_from_grid(
                        grid=self.grid, return_in_2d=False, return_binned=True
                    ),
                    galaxies,
                )
            )
        elif self.use_deflections_y:
            return sum(
                map(
                    lambda g: g.deflections_from_grid(
                        grid=self.grid, return_in_2d=False, return_binned=True
                    ),
                    galaxies,
                )
            )[:, 0]
        elif self.use_deflections_x:
            return sum(
                map(
                    lambda g: g.deflections_from_grid(
                        grid=self.grid, return_in_2d=False, return_binned=True
                    ),
                    galaxies,
                )
            )[:, 1]

    def mask(self, return_in_2d=True):
        if return_in_2d:
            return self.mask_2d
        else:
            return self.mask_1d

    @reshape_array
    def image(self, return_in_2d=True):
        return self.image_1d

    @reshape_array
    def noise_map(self, return_in_2d=True):
        return self.noise_map_1d

    @reshape_array
    def signal_to_noise_map(self, return_in_2d=True):
        return self.signal_to_noise_map_1d



from astropy import cosmology
from scipy.optimize import fsolve
from autolens.model.profiles import mass_profiles as mp

from astropy import constants

import numpy as np


def convert_to_lens_unit(m_nfw, c_nfw, z_nfw, z_source, cosmo=cosmology.LambdaCDM):
    """
    Trying to convert los subhalo with a given mass and concentration
    to lens units.

    m_nfw: m_200 [solar mass]
    c_nfw: concentration
    z_nfw: Redshift of the halo
    z_source: Redshift of the background source

    return kappa_s & rs [arcsec]
    """
    H_l = (cosmo.H0).value  # Hubble constant
    G_l = 4.302e-9  # Mpc M_sun^-1 (km/s)
    J_c = 3.0 * H_l ** 2.0 / (8.0 * np.pi * G_l) / 10.0 ** (9.0)  # M_sun/kpc^(3)
    C_v = 3 * 10 ** 5.0  # light velocity [km/s]
    G_crit = 4.302e-6  # kpc M_sun^-1 (km/s)^2

    r200 = (m_nfw / (200.0 * J_c * (4.0 * np.pi / 3.0))) ** (1.0 / 3.0)  # R_200
    c = c_nfw
    # print(c)
    de_c = 200.0 / 3.0 * (c * c * c / (np.log(1.0 + c) - c / (1.0 + c)))  # delta_c
    rs = r200 / c  # rs
    rhos = J_c * de_c  # rhos
    s_crit = (
        C_v
        * C_v
        * cosmo.angular_diameter_distance(z_source)
        / (
            4.0
            * np.pi
            * cosmo.angular_diameter_distance(z_nfw)
            * cosmo.angular_diameter_distance_z1z2(z_nfw, z_source)
        )
        / (1000.0 * G_crit)
    ).value  # critical density at los halo
    kappa_s = rs * rhos / s_crit
    arctokpc = (
        np.pi / 180.0 / 3600.0 * (cosmo.angular_diameter_distance(z_nfw) * 1000.0).value
    )
    rs_arcsec = rs / arctokpc
    return kappa_s, rs_arcsec


def convert_to_physical_unit(
    kappa_s, rs_arcsec, z_nfw, z_source, cosmo=cosmology.LambdaCDM
):
    """
    Trying to convert NFW lens units to physical units.

    kappa_s: rhos*rs/s_crit
    rs_arcsec: scale radius [arcsec]
    z_nfw: Redshift of the halo
    z_source: Redshift of the background source

    return m200 [solar mass] & concentration: c
    """
    H_l = cosmo.H0.value  # Hubble constant
    G_l = 4.302e-9  # Mpc M_sun^-1 (km/s) # Gravitational constnat
    J_c = (
        3.0 * H_l ** 2.0 / (8.0 * np.pi * G_l) / 10.0 ** (9.0)
    )  # cosmic average density M_sun/kpc^(3)
    C_v = 3 * 10 ** 5.0  # light velocity [km/s]
    G_crit = 4.302e-6  # kpc M_sun^-1 (km/s)^2 # Gravitational constant

    arctokpc = (
        np.pi / 180.0 / 3600.0 * (cosmo.angular_diameter_distance(z_nfw) * 1000.0).value
    )
    rs = rs_arcsec * arctokpc

    s_crit = (
        C_v
        * C_v
        * cosmo.angular_diameter_distance(z_source)
        / (
            4.0
            * np.pi
            * cosmo.angular_diameter_distance(z_nfw)
            * cosmo.angular_diameter_distance_z1z2(z_nfw, z_source)
        )
        / (1000.0 * G_crit)
    ).value  # critical density

    rhos = kappa_s * s_crit / rs  # rho_s

    print(rhos)

    de_c = rhos / J_c  # delta_c

    c = fsolve(solve_c, 10.0, args=(de_c,))[0]  # concentration

    r200 = c * rs  # R_200

    m200 = 200.0 * (4.0 / 3.0 * np.pi) * J_c * r200 ** 3.0  # M_200

    return m200, c


def solve_c(c, de_c):
    """
    Equation need for solving concentration c for a given delta_c
    """
    return 200.0 / 3.0 * (c * c * c / (np.log(1 + c) - c / (1 + c))) - de_c


def convert_to_physical_unit_tNFW(
    kappa_s, rs_arcsec, z_nfw, z_source, cosmo=cosmology.LambdaCDM, rt_arcsec=None
):
    """
    Trying to convert tNFW lens units to physical units.

    kappa_s: rhos*rs/s_crit

    rs_arcsec: scale radius [arcsec]

    rt_arcsec: truncation radius [arcsec]
    if rt_arcsec is not given, then rt is set to be 2.0*R_200

    z_nfw: Redshift of the halo

    z_source: Redshift of the background source

    return m200 [solar mass], m_t [solar_mass], & concentration: c, r200 [kpc]
    """
    H_l = (cosmo.H0).value  # Hubble constant
    G_l = 4.302e-9  # Mpc M_sun^-1 (km/s) # Gravitational constnat
    J_c = (
        cosmo.critical_density(z=0.6).to("solMass / kpc^3").value
    )  # cosmic average density M_sun/kpc^(3)
    C_v = constants.c.to("km / s")
    G_crit = constants.G.to("kpc km2 / (solMass s2)")

    print("Cosmic average density = ", J_c)

    arctokpc = (
        np.pi / 180.0 / 3600.0 * (cosmo.angular_diameter_distance(z_nfw) * 1000.0).value
    )

    print("arcsec to kpc = ", arctokpc)

    print("scae radius in arcsec = ", rs_arcsec)

    rs = rs_arcsec * arctokpc

    print("scale radius in kpc = ", rs)

    s_crit = (
        C_v
        * C_v
        * cosmo.angular_diameter_distance(z_source)
        / (
            4.0
            * np.pi
            * cosmo.angular_diameter_distance(z_nfw)
            * cosmo.angular_diameter_distance_z1z2(z_nfw, z_source)
        )
        / (1000.0 * G_crit)
    )  # critical density

    print("s_crit (native units) = ", s_crit)

    s_crit = s_crit.value

    print("critical density = ", s_crit)

    rhos = kappa_s * s_crit / rs  # rho_s

    print("rho = ", rhos)

    de_c = rhos / J_c  # delta_c

    print("cosmic density = ", J_c)
    print(
        "cosmic density astropy = ", cosmo.critical_density(z=0.0).to("solMass / kpc^3")
    )
    print("delta con = ", de_c)

    c = fsolve(solve_c, 10.0, args=(de_c,))[0]  # concentration

    r200 = c * rs  # R_200

    m200 = 200.0 * (4.0 / 3.0 * np.pi) * J_c * r200 ** 3.0  # M_200

    if rt_arcsec is not None:
        tau = rt_arcsec / rs_arcsec
    else:
        tau = 2.0 * c

    # if rt is given, calculate the tau
    # if rt is not given, set rt = 2.0*r_200, which means tau = 2.0*c

    m_t = (
        m200
        * (tau * tau / (tau * tau + 1.0) ** 2.0)
        * ((tau * tau - 1) * np.log(tau) + tau * np.pi - (tau * tau + 1))
    )

    # m_t is the truncation mass.

    return m200, m_t, c, r200


def convert_to_physical_unit_tNFW_kpc(
    kappa_s, rs_kpc, z_nfw, z_source, cosmo=cosmology.LambdaCDM, rt_kpc=None
):
    """
    Trying to convert tNFW lens units to physical units.

    kappa_s: rhos*rs/s_crit

    rs_arcsec: scale radius [arcsec]

    rt_arcsec: truncation radius [arcsec]
    if rt_arcsec is not given, then rt is set to be 2.0*R_200

    z_nfw: Redshift of the halo

    z_source: Redshift of the background source

    return m200 [solar mass], m_t [solar_mass], & concentration: c, r200 [kpc]
    """
    H_l = (cosmo.H0).value  # Hubble constant
    G_l = 4.302e-9  # Mpc M_sun^-1 (km/s) # Gravitational constnat
    J_c = (
        cosmo.critical_density(z=0.6).to("solMass / kpc^3").value
    )  # cosmic average density M_sun/kpc^(3)
    C_v = constants.c.to("km / s")
    G_crit = constants.G.to("kpc km2 / (solMass s2)")

    print("Cosmic average density = ", J_c)

    arctokpc = (
        np.pi / 180.0 / 3600.0 * (cosmo.angular_diameter_distance(z_nfw) * 1000.0).value
    )

    print("arcsec to kpc = ", arctokpc)

    rs = rs_kpc

    print("scale radius in kpc = ", rs)

    s_crit = (
        C_v
        * C_v
        * cosmo.angular_diameter_distance(z_source)
        / (
            4.0
            * np.pi
            * cosmo.angular_diameter_distance(z_nfw)
            * cosmo.angular_diameter_distance_z1z2(z_nfw, z_source)
        )
        / (1000.0 * G_crit)
    )  # critical density

    print("s_crit (native units) = ", s_crit)

    s_crit = s_crit.value

    print("critical density = ", s_crit)

    rhos = kappa_s * s_crit / rs  # rho_s

    print("rho = ", rhos)

    de_c = rhos / J_c  # delta_c

    print("cosmic density = ", J_c)
    print(
        "cosmic density astropy = ", cosmo.critical_density(z=0.0).to("solMass / kpc^3")
    )
    print("delta con = ", de_c)

    c = fsolve(solve_c, 10.0, args=(de_c,))[0]  # concentration

    r200 = c * rs  # R_200

    m200 = 200.0 * (4.0 / 3.0 * np.pi) * J_c * r200 ** 3.0  # M_200

    if rt_kpc is not None:
        tau = rt_kpc / rs_kpc
    else:
        tau = 2.0 * c

    # if rt is given, calculate the tau
    # if rt is not given, set rt = 2.0*r_200, which means tau = 2.0*c

    m_t = (
        m200
        * (tau * tau / (tau * tau + 1.0) ** 2.0)
        * ((tau * tau - 1) * np.log(tau) + tau * np.pi - (tau * tau + 1))
    )

    # m_t is the truncation mass.

    return m200, m_t, c, r200


def r200_from_rs_arcsec(kappa_s, rs_arcsec, cosmo=cosmology.LambdaCDM):
    """
    Trying to convert tNFW lens units to physical units.

    kappa_s: rhos*rs/s_crit

    rs_arcsec: scale radius [arcsec]

    rt_arcsec: truncation radius [arcsec]
    if rt_arcsec is not given, then rt is set to be 2.0*R_200

    z_nfw: Redshift of the halo

    z_source: Redshift of the background source

    return m200 [solar mass], m_t [solar_mass], & concentration: c, r200 [kpc]
    """

    z_lens = 0.6
    z_source = 2.5
    J_c = (
        cosmo.critical_density(z=z_lens).to("solMass / kpc^3").value
    )  # cosmic average density M_sun/kpc^(3)
    C_v = constants.c.to("km / s")
    G_crit = constants.G.to("kpc km2 / (solMass s2)")

    print("Cosmic average density = ", J_c)

    arctokpc = (
        np.pi / 180.0 / 3600.0 * (cosmo.angular_diameter_distance(z_nfw) * 1000.0).value
    )

    rs = rs_arcsec * arctokpc

    s_crit = (
        C_v
        * C_v
        * cosmo.angular_diameter_distance(z_source)
        / (
            4.0
            * np.pi
            * cosmo.angular_diameter_distance(z_lens)
            * cosmo.angular_diameter_distance_z1z2(z_nfw, z_source)
        )
        / (1000.0 * G_crit)
    )  # critical density

    print(J_c)
    print(s_crit)

    s_crit = s_crit.value
    rhos = kappa_s * s_crit / rs  # rho_s
    de_c = rhos / J_c  # delta_c
    c = fsolve(solve_c, 10.0, args=(de_c,))[0]  # concentration
    r200 = c * rs  # R_200
    print(r200)
    stop
    r200 = r200 / arctokpc
    print(arctokpc)
    print(r200)
    stop

    return r200


z_source = 2.5
z_nfw = 0.6

mass_200, m_t, concentration, r200 = convert_to_physical_unit_tNFW(
    kappa_s=0.5,
    rs_arcsec=5.0,
    z_nfw=0.6,
    z_source=2.5,
    rt_arcsec=10.0,
    cosmo=cosmology.FlatLambdaCDM(H0=70.0, Om0=0.3),
)

print(mass_200, m_t, concentration, r200)

mass_200, m_t, concentration, r200 = convert_to_physical_unit_tNFW_kpc(
    kappa_s=0.5,
    z_nfw=0.6,
    z_source=2.5,
    rs_kpc=3.0,
    rt_kpc=7.0,
    cosmo=cosmology.FlatLambdaCDM(H0=70.0, Om0=0.3),
)

print(mass_200, m_t, concentration, r200 * 6.68549148608755)


r200 = r200_from_rs_arcsec(
    kappa_s=2.0, rs_arcsec=10.0, cosmo=cosmology.FlatLambdaCDM(H0=70.0, Om0=0.3)
)
print(r200)
stop

mass_200, m_t, concentration, r200 = convert_to_physical_unit_tNFW(
    kappa_s=0.002,
    rs_arcsec=30.0,
    z_nfw=0.6,
    z_source=2.5,
    cosmo=cosmology.FlatLambdaCDM(H0=70.0, Om0=0.3),
)

print("{:.4e}".format(mass_200), concentration)

mass_200, m_t, concentration, r200 = convert_to_physical_unit_tNFW(
    kappa_s=0.00001,
    rs_arcsec=5.0,
    z_nfw=0.6,
    z_source=2.5,
    cosmo=cosmology.FlatLambdaCDM(H0=70.0, Om0=0.3),
)

print("{:.4e}".format(mass_200), concentration)

mass_200, m_t, concentration, r200 = convert_to_physical_unit_tNFW(
    kappa_s=0.00001,
    rs_arcsec=1.0,
    z_nfw=0.6,
    z_source=2.5,
    cosmo=cosmology.FlatLambdaCDM(H0=70.0, Om0=0.3),
)

print("{:.4e}".format(mass_200), concentration)

mass_200, m_t, concentration, r200 = convert_to_physical_unit_tNFW(
    kappa_s=0.00001,
    rs_arcsec=30.0,
    z_nfw=0.6,
    z_source=2.5,
    cosmo=cosmology.FlatLambdaCDM(H0=70.0, Om0=0.3),
)

print("{:.4e}".format(mass_200), concentration)



import autofit as af
import autolens as al
from autolens.pipeline import pipeline

# This pipeline fits a strong lens which has two lens galaxies. It is composed of the following 4 phases:

# Phase 1) Fit the light profile of the lens galaxy on the left of the image, at coordinates (0.0", -1.0").

# Phase 2) Fit the light profile of the lens galaxy on the right of the image, at coordinates (0.0", 1.0").

# Phase 3) Use this lens-subtracted image to fit the source galaxy's light. The mass-profiles of the two lens galaxies
#          can use the results of phases 1 and 2 to initialize their priors.

# Phase 4) Fit all relevant parameters simultaneously, using priors from phases 1, 2 and 3.

# Because the pipeline assumes the lens galaxies are at (0.0", -1.0") and (0.0", 1.0"), it is not a general pipeline
# and cannot be applied to any image of a strong lens.


def make_pipeline(phase_folders=None):

    ### SETUP PIPELINE AND PHASE NAMES, TAGS AND PATHS ###

    # We setup the pipeline name using the tagging module. In this case, the pipeline name is not given a tag and
    # will be the string specified below. However, its good practise to use the 'tag.' function below, incase
    # a pipeline does use customized tag names.

    pipeline_name = "pipeline__x2_galaxies"

    # This function uses the phase folders and pipeline name to set up the output directory structure,
    # e.g. 'autolens_workspace/output/phase_folder_1/phase_folder_2/pipeline_name/pipeline_tag/phase_name/phase_tag/'

    phase_folders.append(pipeline_name)

    ### PHASE 1 ###

    # The left-hand galaxy is at (0.0", -1.0"), so we're going to use a small circular mask centred on its location to
    # fit its light profile. Its important that light from the other lens galaxy and source galaxy don't contaminate
    # our fit.

    def mask_function(image):
        return al.Mask.circular(
            shape=image.shape,
            pixel_scale=image.pixel_scale,
            radius_arcsec=0.5,
            centre=(0.0, -1.0),
        )

    class LeftLensPhase(al.PhaseImaging):
        def customize_priors(self, results):

            # Let's restrict the priors on the centres around the pixel we know the galaxy's light centre peaks.

            self.galaxies.left_lens.light.centre_0 = af.GaussianPrior(
                mean=0.0, sigma=0.05
            )

            self.galaxies.left_lens.light.centre_1 = af.GaussianPrior(
                mean=-1.0, sigma=0.05
            )

            # Given we are only fitting the very central region of the lens galaxy, we don't want to let a parameter
            # like the Sersic index vary. Lets fix it to 4.0.

            self.galaxies.left_lens.light.sersic_index = 4.0

    phase1 = LeftLensPhase(
        phase_name="phase_1__left_lens_light",
        phase_folders=phase_folders,
        galaxies=dict(
            left_lens=al.GalaxyModel(
                redshift=0.5, light=al.light_profiles.EllipticalSersic
            )
        ),
        mask_function=mask_function,
        optimizer_class=af.MultiNest,
    )

    phase1.optimizer.const_efficiency_mode = True
    phase1.optimizer.n_live_points = 30
    phase1.optimizer.sampling_efficiency = 0.5

    ### PHASE 2 ###

    # Now do the exact same with the lens galaxy on the right at (0.0", 1.0")

    def mask_function(image):
        return al.Mask.circular(
            image.shape,
            pixel_scale=image.pixel_scale,
            radius_arcsec=0.5,
            centre=(0.0, 1.0),
        )

    class RightLensPhase(al.PhaseImaging):
        def customize_priors(self, results):

            self.galaxies.right_lens.light.centre_0 = af.GaussianPrior(
                mean=0.0, sigma=0.05
            )

            self.galaxies.right_lens.light.centre_1 = af.GaussianPrior(
                mean=1.0, sigma=0.05
            )

            self.galaxies.right_lens.light.sersic_index = 4.0

    phase2 = RightLensPhase(
        phase_name="phase_2__right_lens_light",
        phase_folders=phase_folders,
        galaxies=dict(
            right_lens=al.GalaxyModel(
                redshift=0.5, light=al.light_profiles.EllipticalSersic
            )
        ),
        mask_function=mask_function,
        optimizer_class=af.MultiNest,
    )

    phase2.optimizer.const_efficiency_mode = True
    phase2.optimizer.n_live_points = 30
    phase2.optimizer.sampling_efficiency = 0.5

    ### PHASE 3 ###

    # In the next phase, we fit the source of the lens subtracted image.

    class LensSubtractedPhase(al.PhaseImaging):

        # To modify the image, we want to subtract both the left-hand and right-hand lens galaxies. To do this, we need
        # to subtract the unmasked model image of both galaxies!

        def modify_image(self, image, results):

            phase_1_results = results.from_phase("phase_1__left_lens_light")
            phase_2_results = results.from_phase("phase_2__right_lens_light")

            return (
                image
                - phase_1_results.unmasked_model_image_of_planes[0]
                - phase_2_results.unmasked_model_image_of_planes[0]
            )

        def customize_priors(self, results):

            phase_1_results = results.from_phase("phase_1__left_lens_light")
            phase_2_results = results.from_phase("phase_2__right_lens_light")

            # We're going to link the centres of the light profiles computed above to the centre of the lens galaxy
            # mass-profiles in this phase. Because the centres of the mass profiles were fixed in phases 1 and 2,
            # linking them using the 'variable' attribute means that they stay constant (which for now, is what we want).

            self.galaxies.left_lens.mass.centre_0 = (
                phase_1_results.variable.galaxies.left_lens.light.centre_0
            )

            self.galaxies.left_lens.mass.centre_1 = (
                phase_1_results.variable.galaxies.left_lens.light.centre_1
            )

            self.galaxies.right_lens.mass.centre_0 = (
                phase_2_results.variable.galaxies.right_lens.light.centre_0
            )

            self.galaxies.right_lens.mass.centre_1 = (
                phase_2_results.variable.galaxies.right_lens.light.centre_1
            )

    phase3 = LensSubtractedPhase(
        phase_name="phase_3__lens_x2_sie__source_exp",
        phase_folders=phase_folders,
        galaxies=dict(
            left_lens=al.GalaxyModel(
                redshift=0.5, mass=al.mass_profiles.EllipticalIsothermal
            ),
            right_lens=al.GalaxyModel(
                redshift=0.5, mass=al.mass_profiles.EllipticalIsothermal
            ),
            source=al.GalaxyModel(
                redshift=1.0, light=al.light_profiles.EllipticalExponential
            ),
        ),
        optimizer_class=af.MultiNest,
    )

    phase3.optimizer.const_efficiency_mode = True
    phase3.optimizer.n_live_points = 50
    phase3.optimizer.sampling_efficiency = 0.5

    ### PHASE 4 ###

    # In phase 4, we'll fit both lens galaxy's light and mass profiles, as well as the source-galaxy, simultaneously.

    class FitAllPhase(al.PhaseImaging):
        def customize_priors(self, results):

            phase_1_results = results.from_phase("phase_1__left_lens_light")
            phase_2_results = results.from_phase("phase_2__right_lens_light")
            phase_3_results = results.from_phase("phase_3__lens_x2_sie__source_exp")

            # Results are split over multiple phases, so we setup the light and mass profiles of each lens separately.

            self.galaxies.left_lens.light = (
                phase_1_results.variable.galaxies.left_lens.light
            )

            self.galaxies.left_lens.mass = (
                phase_3_results.variable.galaxies.left_lens.mass
            )

            self.galaxies.right_lens.light = (
                phase_2_results.variable.galaxies.right_lens.light
            )

            self.galaxies.right_lens.mass = (
                phase_3_results.variable.galaxies.right_lens.mass
            )

            # When we pass a a 'variable' galaxy from a previous phase, parameters fixed to constants remain constant.
            # Because centre_0 and centre_1 of the mass profile were fixed to constants in phase 3, they're still
            # constants after the line after. We need to therefore manually over-ride their priors.

            self.galaxies.left_lens.mass.centre_0 = (
                phase_3_results.variable.galaxies.left_lens.mass.centre_0
            )

            self.galaxies.left_lens.mass.centre_1 = (
                phase_3_results.variable.galaxies.left_lens.mass.centre_1
            )

            self.galaxies.right_lens.mass.centre_0 = (
                phase_3_results.variable.galaxies.right_lens.mass.centre_0
            )

            self.galaxies.right_lens.mass.centre_1 = (
                phase_3_results.variable.galaxies.right_lens.mass.centre_1
            )

            # We also want the Sersic index's to be free parameters now, so lets change it from a constant to a
            # variable.

            self.galaxies.left_lens.light.sersic_index = af.GaussianPrior(
                mean=4.0, sigma=2.0
            )

            self.galaxies.right_lens.light.sersic_index = af.GaussianPrior(
                mean=4.0, sigma=2.0
            )

            # Things are much simpler for the source galaxies - just like them together!

            self.galaxies.source = phase_3_results.variable.galaxies.source

    phase4 = FitAllPhase(
        phase_name="phase_4__lens_x2_sersic_sie__source_exp",
        phase_folders=phase_folders,
        galaxies=dict(
            left_lens=al.GalaxyModel(
                redshift=0.5,
                light=al.light_profiles.EllipticalSersic,
                mass=al.mass_profiles.EllipticalIsothermal,
            ),
            right_lens=al.GalaxyModel(
                redshift=0.5,
                light=al.light_profiles.EllipticalSersic,
                mass=al.mass_profiles.EllipticalIsothermal,
            ),
            source=al.GalaxyModel(
                redshift=1.0, light=al.light_profiles.EllipticalExponential
            ),
        ),
        optimizer_class=af.MultiNest,
    )

    phase4.optimizer.const_efficiency_mode = True
    phase4.optimizer.n_live_points = 60
    phase4.optimizer.sampling_efficiency = 0.5

    return pipeline.PipelineImaging(pipeline_name, phase1, phase2, phase3, phase4)



from astropy import cosmology
from astropy import constants
from astropy import units
from autolens.model.profiles import mass_profiles as mp

import numpy as np

cosmo = cosmology.FlatLambdaCDM(H0=70.0, Om0=0.3)

z_nfw = 0.6
z_source = 2.5

arc_per_kpc = cosmo.arcsec_per_kpc_proper(z=z_nfw)
H_l = cosmo.H0  # Hubble constant
G_l = constants.G
J_c = cosmo.critical_density(z=z_nfw)  # cosmic average density M_sun/kpc^(3)
C_v = constants.c

print("arc per kpc = ", arc_per_kpc)
print("H0 = ", H_l)
print("G = ", G_l)
print("Critical Density = ", J_c)
print("Speed of light = ", C_v)

s_crit = (
    C_v ** 2
    * cosmo.angular_diameter_distance(z_source)
    / (
        4.0
        * G_l
        * np.pi
        * cosmo.angular_diameter_distance(z_nfw)
        * cosmo.angular_diameter_distance_z1z2(z_nfw, z_source)
    )
)  # critical density

print()

print(s_crit.to("solMass / kpc^2"))
print(J_c.to("solMass / kpc^3"))
print(s_crit.to("solMass / kpc^2") / arc_per_kpc ** 2.0)
print(J_c.to("solMass / kpc^3") / arc_per_kpc ** 3.0)

stop

print()

mass_200 = 10 ** 6
concentration = 11.5 * ((mass_200 / 10 ** 10) + (mass_200 / 10 ** 10) ** 2.0) ** -0.05
print("Concentration 10**6 ", concentration)

mass_200 = 10 ** 8
concentration = 11.5 * ((mass_200 / 10 ** 10) + (mass_200 / 10 ** 10) ** 2.0) ** -0.05
print("Concentration 10**8 ", concentration)

mass_200 = 10 ** 10
concentration = 11.5 * ((mass_200 / 10 ** 10) + (mass_200 / 10 ** 10) ** 2.0) ** -0.05
print("Concentration 10**10 ", concentration)

nfw_challenge = al.mass_profiles.SphericalTruncatedNFWChallenge(
    kappa_s=0.003, scale_radius=0.1
)
mass_at_200 = nfw_challenge.mass_at_200_for_units(
    critical_surface_density=1940654909.4133248,
    cosmic_average_density=262.30319684750657,
)
radius_at_200 = nfw_challenge.radius_at_200_for_units(
    critical_surface_density=1940654909.4133248,
    cosmic_average_density=262.30319684750657,
)
concentration = nfw_challenge.concentration_for_units(
    critical_surface_density=1940654909.4133248,
    cosmic_average_density=262.30319684750657,
)
mass_at_truncation_radius = nfw_challenge.mass_at_truncation_radius(
    critical_surface_density=1940654909.4133248,
    cosmic_average_density=262.30319684750657,
)

print(mass_200, radius_at_200, concentration, mass_at_truncation_radius)
stop

print()
print("mass = 10^6")
nfw_challenge = al.mass_profiles.SphericalTruncatedNFWChallenge(
    kappa_s=0.003, scale_radius=0.1
)
mass_at_200 = nfw_challenge.mass_at_200_for_units(
    critical_surface_density=1940654909.4133248,
    cosmic_average_density=262.30319684750657,
)
concentration = nfw_challenge.concentration_for_units(
    critical_surface_density=1940654909.4133248,
    cosmic_average_density=262.30319684750657,
)


print("{:.4e}".format(mass_at_200[0]))
print(concentration)

print()
print("mass = 10^8")
nfw_challenge = al.mass_profiles.SphericalTruncatedNFWChallenge(
    kappa_s=0.01, scale_radius=0.5
)
mass_at_200 = nfw_challenge.mass_at_200_for_units(
    critical_surface_density=1940654909.4133248,
    cosmic_average_density=262.30319684750657,
)
concentration = nfw_challenge.concentration_for_units(
    critical_surface_density=1940654909.4133248,
    cosmic_average_density=262.30319684750657,
)


print("{:.4e}".format(mass_at_200[0]))
print(concentration)

print()
print("mass = 10^10")
nfw_challenge = al.mass_profiles.SphericalTruncatedNFWChallenge(
    kappa_s=0.03, scale_radius=3.0
)
mass_at_200 = nfw_challenge.mass_at_200_for_units(
    critical_surface_density=1940654909.4133248,
    cosmic_average_density=262.30319684750657,
)
concentration = nfw_challenge.concentration_for_units(
    critical_surface_density=1940654909.4133248,
    cosmic_average_density=262.30319684750657,
)


print("{:.4e}".format(mass_at_200[0]))
print(concentration)



from autolens.array.mapping_util import array_mapping_util
from autolens.array import scaled_array
from autolens.model.inversion.util import mapper_util

import numpy as np


class Mapper(object):
    def __init__(self, pixels, grid, pixelization_grid, hyper_image=None):
        """ Abstract base class representing a mapper, which maps unmasked pixels on a masked 2D array (in the form of \
        a grid, see the *hyper_galaxies.array.grid* module) to discretized pixels in a pixelization.

        1D arrays are used to represent these mappings, for example between the different grid in a grid \
        (e.g. the / sub grid). This follows the syntax grid_to_grid, whereby the index of a value on one grid \
        equals that of another grid, for example:

        - image_to_pix[2] = 1  tells us that the 3rd pixel on a grid maps to the 2nd pixel of a pixelization.
        - sub_to_pix4] = 2  tells us that the 5th sub-pixel of a sub-grid maps to the 3rd pixel of a pixelization.
        - pix_to_image[2] = 5 tells us that the 3rd pixel of a pixelization maps to the 6th (unmasked) pixel of a \
                            grid.

        Parameters
        ----------
        pixels : int
            The number of pixels in the mapper's pixelization.
        grid: gridStack
            A stack of grid's which are mapped to the pixelization (includes an and sub grid).
        border : grid.GridBorder
            The border of the grid's grid.
        hyper_image : ndarray
            A pre-computed hyper_galaxies-image of the image the mapper is expected to reconstruct, used for adaptive analysis.
        """
        self.pixels = pixels
        self.grid = grid
        self.pixelization_grid = pixelization_grid
        self.hyper_image = hyper_image

    @property
    def mapping_matrix(self):
        """The mapping_util matrix is a matrix representing the mapping_util between every unmasked pixel of a grid and \
        the pixels of a pixelization. Non-zero entries signify a mapping_util, whereas zeros signify no mapping_util.

        For example, if the grid has 5 pixels and the pixelization 3 pixels, with the following mappings:

        pixel 0 -> pixelization pixel 0
        pixel 1 -> pixelization pixel 0
        pixel 2 -> pixelization pixel 1
        pixel 3 -> pixelization pixel 1
        pixel 4 -> pixelization pixel 2

        The mapping_util matrix (which is of dimensions regular_pixels x pixelization_pixels) would appear as follows:

        [1, 0, 0] [0->0]
        [1, 0, 0] [1->0]
        [0, 1, 0] [2->1]
        [0, 1, 0] [3->1]
        [0, 0, 1] [4->2]

        The mapping_util matrix is in fact built using the sub-grid of the grid, whereby each pixel is \
        divided into a grid of sub-pixels which are all paired to pixels in the pixelization. The entires \
        in the mapping_util matrix now become fractional values dependent on the sub-grid size. For example, for a 2x2 \
        sub-grid in each pixel (which means the fraction value is 1.0/(2.0^2) = 0.25, if we have the following mappings:

        pixel 0 -> sub pixel 0 -> pixelization pixel 0
        pixel 0 -> sub pixel 1 -> pixelization pixel 1
        pixel 0 -> sub pixel 2 -> pixelization pixel 1
        pixel 0 -> sub pixel 3 -> pixelization pixel 1
        pixel 1 -> sub pixel 0 -> pixelization pixel 1
        pixel 1 -> sub pixel 1 -> pixelization pixel 1
        pixel 1 -> sub pixel 2 -> pixelization pixel 1
        pixel 1 -> sub pixel 3 -> pixelization pixel 1
        pixel 2 -> sub pixel 0 -> pixelization pixel 2
        pixel 2 -> sub pixel 1 -> pixelization pixel 2
        pixel 2 -> sub pixel 2 -> pixelization pixel 3
        pixel 2 -> sub pixel 3 -> pixelization pixel 3

        The mapping_util matrix (which is still of dimensions regular_pixels x source_pixels) would appear as follows:

        [0.25, 0.75, 0.0, 0.0] [1 sub-pixel maps to pixel 0, 3 map to pixel 1]
        [ 0.0,  1.0, 0.0, 0.0] [All sub-pixels map to pixel 1]
        [ 0.0,  0.0, 0.5, 0.5] [2 sub-pixels map to pixel 2, 2 map to pixel 3]
        """
        return mapper_util.mapping_matrix_from_sub_mask_1d_index_to_pixelization_1d_index(
            sub_mask_1d_index_to_pixelization_1d_index=self.sub_mask_1d_index_to_pixelization_1d_index,
            pixels=self.pixels,
            total_mask_pixels=self.grid.mask.pixels_in_mask,
            sub_mask_1d_index_to_mask_1d_index=self.grid.sub_mask_1d_index_to_mask_1d_index,
            sub_grid_fraction=self.grid.sub_grid_fraction,
        )

    @property
    def sub_mask_1d_index_to_pixelization_1d_index(self):
        raise NotImplementedError("sub_to_pixelization should be overridden")

    @property
    def pixelization_1d_index_to_all_sub_mask_1d_indexes(self):
        """Compute the mappings between a pixelization's pixels and the unmasked sub-grid pixels. These mappings \
        are determined after the grid is used to determine the pixelization.

        The pixelization's pixels map to different number of sub-grid pixels, thus a list of lists is used to \
        represent these mappings"""

        pixelization_1d_index_to_all_sub_mask_1d_indexes = [
            [] for _ in range(self.pixels)
        ]

        for mask_1d_index, pixelization_1d_index in enumerate(
            self.sub_mask_1d_index_to_pixelization_1d_index
        ):
            pixelization_1d_index_to_all_sub_mask_1d_indexes[
                pixelization_1d_index
            ].append(mask_1d_index)

        return pixelization_1d_index_to_all_sub_mask_1d_indexes


class RectangularMapper(Mapper):
    def __init__(
        self, pixels, grid, pixelization_grid, shape, geometry, hyper_image=None
    ):
        """ Class representing a rectangular mapper, which maps unmasked pixels on a masked 2D array (in the form of \
        a grid, see the *hyper_galaxies.array.grid* module) to pixels discretized on a rectangular grid.

        The and uniform geometry of the rectangular grid is used to perform efficient pixel pairings.

        Parameters
        ----------
        pixels : int
            The number of pixels in the rectangular pixelization (y_pixels*x_pixels).
        grid : gridStack
            A stack of grid describing the observed image's pixel coordinates (e.g. an image-grid, sub-grid, etc.).
        border : grid.GridBorder
            The border of the grid's grid.
        shape : (int, int)
            The dimensions of the rectangular grid of pixels (y_pixels, x_pixel)
        geometry : pixelization.Rectangular.Geometry
            The geometry (e.g. y / x edge locations, pixel-scales) of the rectangular pixelization.
        """
        self.shape = shape
        self.geometry = geometry
        super(RectangularMapper, self).__init__(
            pixels=pixels,
            grid=grid,
            pixelization_grid=pixelization_grid,
            hyper_image=hyper_image,
        )

    @property
    def is_image_plane_pixelization(self):
        return False

    @property
    def sub_mask_1d_index_to_pixelization_1d_index(self):
        """The 1D index mappings between the sub grid's pixels and rectangular pixelization's pixels"""
        return self.geometry.grid_arcsec_1d_to_grid_pixel_indexes_1d(
            grid_arcsec=self.grid
        )

    def reconstructed_pixelization_from_solution_vector(self, solution_vector):
        """Given the solution vector of an inversion (see *inversions.Inversion*), determine the reconstructed \
        pixelization of the rectangular pixelization by using the mapper."""
        recon = array_mapping_util.sub_array_2d_from_sub_array_1d_mask_and_sub_grid_size(
            sub_array_1d=solution_vector,
            mask=np.full(fill_value=False, shape=self.shape),
            sub_grid_size=1,
        )
        return scaled_array.ScaledRectangularPixelArray(
            array=recon,
            pixel_scales=self.geometry.pixel_scales,
            origin=self.geometry.origin,
        )


class VoronoiMapper(Mapper):
    def __init__(
        self, pixels, grid, pixelization_grid, voronoi, geometry, hyper_image=None
    ):
        """Class representing a Voronoi mapper, which maps unmasked pixels on a masked 2D array (in the form of \
        a grid, see the *hyper_galaxies.array.grid* module) to pixels discretized on a Voronoi grid.

        The irand non-uniform geometry of the Voronoi grid means efficient pixel pairings requires knowledge \
        of how different grid map to one another.

        Parameters
        ----------
        pixels : int
            The number of pixels in the Voronoi pixelization.
        grid : gridStack
            A stack of grid describing the observed image's pixel coordinates (e.g. an image-grid, sub-grid, etc.).
        border : grid.GridBorder
            The border of the grid's grid.
        voronoi : scipy.spatial.Voronoi
            Class storing the Voronoi grid's geometry.
        geometry : pixelization.Voronoi.Geometry
            The geometry (e.g. y / x edge locations, pixel-scales) of the Vornoi pixelization.
        hyper_image : ndarray
            A pre-computed hyper_galaxies-image of the image the mapper is expected to reconstruct, used for adaptive analysis.
        """
        self.voronoi = voronoi
        self.geometry = geometry
        super(VoronoiMapper, self).__init__(
            pixels=pixels,
            grid=grid,
            pixelization_grid=pixelization_grid,
            hyper_image=hyper_image,
        )

    @property
    def is_image_plane_pixelization(self):
        return True

    @property
    def sub_mask_1d_index_to_pixelization_1d_index(self):
        """  The 1D index mappings between the sub pixels and Voronoi pixelization pixels. """
        return mapper_util.voronoi_sub_mask_1d_index_to_pixeliztion_1d_index_from_grids_and_geometry(
            grid=self.grid,
            mask_1d_index_to_nearest_pixelization_1d_index=self.pixelization_grid.mask_1d_index_to_nearest_pixelization_1d_index,
            sub_mask_1d_index_to_mask_1d_index=self.grid.sub_mask_1d_index_to_mask_1d_index,
            pixel_centres=self.geometry.pixel_centres,
            pixel_neighbors=self.geometry.pixel_neighbors,
            pixel_neighbors_size=self.geometry.pixel_neighbors_size,
        ).astype(
            "int"
        )



import autolens as al
from autolens.model.inversion import inversions as inv
from autolens.model.inversion import pixelizations as pix
from autolens.model.inversion import regularization as reg
from autolens.model.inversion.plotters import inversion_plotters
from autolens.model.inversion.plotters import mapper_plotters

# We've covered mappers, which, if I haven't emphasised it enough yet, map things. Now, we're going to look at how we
# can use these mappers (which map things) to reconstruct the source galaxy - I hope you're excited!

# Lets simulate the same lens as before. The source is simple - which kind of defeats the point of using a pixel-grid
# to reconstruct it, but we'll make things more complex later on!
def simulate():

    psf = al.PSF.from_gaussian(shape=(11, 11), sigma=0.05, pixel_scale=0.05)

    grid = al.Grid.from_shape_pixel_scale_and_sub_grid_size(
        shape=(180, 180), pixel_scale=0.05
    )

    lens_galaxy = al.Galaxy(
        redshift=0.5,
        mass=al.mass_profiles.EllipticalIsothermal(
            centre=(0.0, 0.0), axis_ratio=0.8, phi=135.0, einstein_radius=1.6
        ),
    )

    source_galaxy = al.Galaxy(
        redshift=1.0,
        light=al.light_profiles.EllipticalSersic(
            centre=(0.0, 0.0),
            axis_ratio=0.8,
            phi=90.0,
            intensity=0.2,
            effective_radius=1.0,
            sersic_index=1.5,
        ),
    )

    tracer = al.Tracer.from_galaxies(galaxies=[lens_galaxy, source_galaxy])

    return al.SimulatedCCDData.from_tracer_grid_and_exposure_arrays(
        tracer=tracer,
        grid=grid,
        pixel_scale=0.05,
        exposure_time=300.0,
        psf=psf,
        background_sky_level=0.1,
        add_noise=True,
    )


# Now, lets simulate the source, mask it, and use a plot to check the masking is appropriate.
ccd_data = simulate()

mask = al.Mask.circular_annular(
    shape=ccd_data.shape,
    pixel_scale=ccd_data.pixel_scale,
    inner_radius_arcsec=1.0,
    outer_radius_arcsec=2.2,
)

al.ccd_plotters.plot_image(ccd_data=ccd_data, mask=mask)

# Next, lets set this image up as lens data and setup a tracer using the input lens galaxy model (we don't need
# to provide the source's light profile, as we're using a mapper to reconstruct it).
lens_data = al.LensData(ccd_data=ccd_data, mask=mask, sub_grid_size=1)

lens_galaxy = al.Galaxy(
    redshift=0.5,
    mass=al.mass_profiles.EllipticalIsothermal(
        centre=(0.0, 0.0), axis_ratio=0.8, phi=135.0, einstein_radius=1.6
    ),
)

tracer = al.Tracer.from_galaxies(galaxies=[lens_galaxy, al.Galaxy(redshift=1.0)])

source_plane_grid = tracer.traced_grids_of_planes_from_grid(grid=lens_data.grid)[1]

# We'll use another rectangular pixelization and mapper to perform the reconstruction
rectangular = pix.Rectangular(shape=(25, 25))

mapper = rectangular.mapper_from_grid_and_pixelization_grid(grid=source_plane_grid)

mapper_plotters.plot_image_and_mapper(
    ccd_data=ccd_data, mask=mask, mapper=mapper, should_plot_grid=True
)

# And now, finally, we're going to use our mapper to invert the image using the 'inversions' module, which is imported
# as 'inv'. I'll explain how this works in a second - but lets just go ahead and perform the inversion first.
# (Ignore the 'regularization' input below for now, we'll cover this in the next tutorial).
inversion = inv.Inversion.from_data_1d_mapper_and_regularization(
    image_1d=lens_data.image_1d,
    noise_map_1d=lens_data.noise_map_1d,
    convolver=lens_data.convolver,
    mapper=mapper,
    regularization=reg.Constant(coefficient=1.0),
)

# Our inversion has a reconstructed image and pixeilzation, whcih we can plot using an inversion plotter
inversion_plotters.plot_reconstructed_image(inversion=inversion, mask=mask)

inversion_plotters.plot_pixelization_values(inversion=inversion, should_plot_grid=True)

# And there we have it, we've successfully reconstructed, or, *inverted*, our source using the mapper's rectangular
# grid. Whilst this source was simple (a blob of light in the centre of the source-plane), inversions come into their
# own when fitting sources with complex morphologies. Infact, given we're having so much fun inverting things, lets
# simulate a really complex source and invert it!


def simulate_complex_source():

    psf = al.PSF.from_gaussian(shape=(11, 11), sigma=0.05, pixel_scale=0.05)

    grid = al.Grid.from_shape_pixel_scale_and_sub_grid_size(
        shape=(180, 180), pixel_scale=0.05
    )

    lens_galaxy = al.Galaxy(
        redshift=0.5,
        mass=al.mass_profiles.EllipticalIsothermal(
            centre=(0.0, 0.0), axis_ratio=0.8, phi=135.0, einstein_radius=1.6
        ),
    )

    source_galaxy_0 = al.Galaxy(
        redshift=1.0,
        light=al.light_profiles.EllipticalSersic(
            centre=(0.1, 0.1),
            axis_ratio=0.8,
            phi=90.0,
            intensity=0.2,
            effective_radius=1.0,
            sersic_index=1.5,
        ),
    )

    source_galaxy_1 = al.Galaxy(
        redshift=1.0,
        light=al.light_profiles.EllipticalSersic(
            centre=(-0.25, 0.25),
            axis_ratio=0.7,
            phi=45.0,
            intensity=0.1,
            effective_radius=0.2,
            sersic_index=3.0,
        ),
    )

    source_galaxy_2 = al.Galaxy(
        redshift=1.0,
        light=al.light_profiles.EllipticalSersic(
            centre=(0.45, -0.35),
            axis_ratio=0.6,
            phi=90.0,
            intensity=0.03,
            effective_radius=0.3,
            sersic_index=3.5,
        ),
    )

    source_galaxy_3 = al.Galaxy(
        redshift=1.0,
        light=al.light_profiles.EllipticalSersic(
            centre=(-0.05, -0.0),
            axis_ratio=0.9,
            phi=140.0,
            intensity=0.03,
            effective_radius=0.1,
            sersic_index=4.0,
        ),
    )

    source_galaxy_4 = al.Galaxy(
        redshift=1.0,
        light=al.light_profiles.EllipticalSersic(
            centre=(0.85, -0.85),
            axis_ratio=0.6,
            phi=90.0,
            intensity=0.03,
            effective_radius=0.3,
            sersic_index=3.5,
        ),
    )

    source_galaxy_5 = al.Galaxy(
        redshift=1.0,
        light=al.light_profiles.EllipticalSersic(
            centre=(-0.75, -0.1),
            axis_ratio=0.9,
            phi=140.0,
            intensity=0.03,
            effective_radius=0.1,
            sersic_index=4.0,
        ),
    )

    tracer = al.Tracer.from_galaxies(
        galaxies=[
            lens_galaxy,
            source_galaxy_0,
            source_galaxy_1,
            source_galaxy_2,
            source_galaxy_3,
            source_galaxy_4,
            source_galaxy_5,
        ]
    )

    return al.SimulatedCCDData.from_tracer_grid_and_exposure_arrays(
        tracer=tracer,
        grid=grid,
        pixel_scale=0.05,
        exposure_time=300.0,
        psf=psf,
        background_sky_level=0.1,
        add_noise=True,
    )


# This code is doing all the the same as above (setup the image, galaxies, tracer, mapper, ec.),
# but I have made the mask slightly larger for this source.
ccd_data = simulate_complex_source()

mask = al.Mask.circular_annular(
    shape=ccd_data.shape,
    pixel_scale=ccd_data.pixel_scale,
    inner_radius_arcsec=0.1,
    outer_radius_arcsec=3.2,
)

al.ccd_plotters.plot_image(ccd_data=ccd_data, mask=mask)

lens_data = al.LensData(ccd_data=ccd_data, mask=mask, sub_grid_size=1)

tracer = al.Tracer.from_galaxies(galaxies=[lens_galaxy, al.Galaxy(redshift=1.0)])

source_plane_grid = tracer.traced_grids_of_planes_from_grid(grid=lens_data.grid)[1]

mapper = rectangular.mapper_from_grid_and_pixelization_grid(grid=source_plane_grid)

inversion = inv.Inversion.from_data_1d_mapper_and_regularization(
    image_1d=lens_data.image_1d,
    noise_map_1d=lens_data.noise_map_1d,
    convolver=lens_data.convolver,
    mapper=mapper,
    regularization=reg.Constant(coefficient=1.0),
)

# Lets inspect the complex source reconstruction.
inversion_plotters.plot_reconstructed_image(inversion=inversion, mask=mask)

inversion_plotters.plot_pixelization_values(inversion=inversion, should_plot_grid=True)

# Pretty great, huh? If you ran the complex source pipeline, you'll remember that getting a model image that looked
# that good simply *was not possible*. With an inversion, we can do it with ease - and without fitting 30+ parameters!

# Lets discuss how an inversion actually works. The explanation I give below is overly-simplified. I'm avoiding the
# technical details of how an inversion *actually* works. To be good at lens modeling you don't need to understand the
# nitty-gritty details of linear inversions, you just need an instinct for how to use them as a tool to model lenses.

# Nevertheless, I know a lot of you hate 'black-boxes', or have an interest in linear algrebra. If you're that way
# inclined, then checkout the documentation of the autolens source code for more information. In particular,
# you should look at the following functions:

# inversions.mappers.mapping_matrix
# inversions.convolution.convolve_mapping_matrix
# inversions.regularization.Regularization
# inversions.inversions.Inversion

# To begin, lets consider some random mappings between our mapper's source-pixels and the image.
mapper_plotters.plot_image_and_mapper(
    ccd_data=ccd_data,
    mapper=mapper,
    mask=mask,
    should_plot_grid=True,
    source_pixels=[[445], [285], [313], [132], [11]],
)

# These mappings are known before the inversion, which means pre-inversion we know two key pieces of information:

# 1) The mappings between every source-pixel and sets of image-pixels.
# 2) The flux values in every observed image-pixel, which are the values we want to fit successfully.

# It turns out that with these two pieces of information we can linearly solve for the set of source-pixel fluxes that
# best-fit (e.g. maximize the likelihood of) our observed image. Essentially, we set up the mapping between source and
# image pixels as a large matrix and solve for the source-pixel fluxes in an analogous fashion to how you would
# solve a set of simultaneous linear equations. This process is called a 'linear inversion'.

# There are three more things about a linear inversion that are worth knowing:

# 1) We've discussed the image sub-grid before, which splits each image-pixel into a sub-pixel. If a sub-grid is used,
#    it is the mapping between every sub-pixel and source-pixel that is computed and used to perform the inversion.
#    This prevents aliasing effects degrading the image reconstruction, and, as a rule of thumb, I would
#    suggest you use sub-gridding of degree 2x2.

# 2) When fitting images using light profiles we discussed how a 'model_image' was generated by blurring them with
#    the data's PSF. A similar blurring operation is incorporated into the inversion, such that the reconstructed
#    image and source fully account for the telescope optics and effect of the PSF.

# 3) The inversion's solution is regularized. But wait, that's what we'll cover in the next tutorial!

# Finally, let me show you how easy it is to fit an image with an inversion using the fitting module. Instead of giving
# the source galaxy a light profile, we give it a pixelization and regularization, and pass it to a tracer.
source_galaxy = al.Galaxy(
    redshift=1.0,
    pixelization=pix.Rectangular(shape=(40, 40)),
    regularization=reg.Constant(coefficient=1.0),
)

tracer = al.Tracer.from_galaxies(galaxies=[lens_galaxy, source_galaxy])

# Then, like before, we call on the fitting module to perform the fit to the lensing image. Indeed, we see
# some pretty good looking residuals - we're certainly fitting the lensed source accurately!
fit = al.LensDataFit.for_data_and_tracer(lens_data=lens_data, tracer=tracer)

al.lens_fit_plotters.plot_fit_subplot(
    fit=fit, should_plot_mask=True, extract_array_from_mask=True, zoom_around_mask=True
)

# And, we're done, here are a few questions to get you thinking about inversions:

# 1) The inversion provides the best-fit solution to the observed image. Is there a problem with seeking the 'best-fit'?
#    Is there a risk that we're going to fit other things in the image than just the lensed source galaxy? What happens
#    if you reduce the regularization 'coefficient' above to zero?

# 2) The exterior pixels in the rectangular grid have no image-pixels in them. However, they are still given a
#    reconstructed flux. If this value isn't' coming from a mapping_util to an image-pixel, where is it be coming from?
